"""
Buffer jobs and tasks into the queue.

Also known as "late materialization" of jobs and tasks,
this method finds datasets that are not fully materialized and
buffers a few more jobs and tasks into existence so they can be queued.

Initial delay: rand(10 minutes)
Periodic delay: 10 minutes
"""

import os
import logging
import random
import time
import asyncio

from tornado.ioloop import IOLoop
from rest_tools.client import RestClient

from iceprod.core.parser import ExpParser
from iceprod.core.resources import Resources
from iceprod.server.priority import Priority

logger = logging.getLogger('buffer_jobs_tasks')

def buffer_jobs_tasks(module):
    """
    Initial entrypoint.

    Args:
        module (:py:class:`iceprod.server.modules.schedule`): schedule module
    """
    # set up rest client
    if 'materialization' in module.cfg and 'url' in module.cfg['materialization']:
        url = module.cfg['materialization']['url']
    else:
        raise Exception('no materialization url')

    if ('rest_api' in module.cfg and 'url' in module.cfg['rest_api']
        and 'auth_key' in module.cfg['rest_api']):
        auth_key = module.cfg['rest_api']['auth_key']
    else:
        raise Exception('no auth key')
    rest_client = RestClient(url, auth_key)

    IOLoop.current().call_later(random.randint(10,60*10), run, rest_client)

async def run(rest_client, only_dataset=None, num=1000, run_once=False, debug=False):
    """
    Actual runtime / loop.

    Args:
        rest_client (:py:class:`iceprod.core.rest_client.Client`): rest client
        only_dataset (str): dataset_id if we should only buffer a single dataset
        num (int): max number of jobs per dataset to buffer
        run_once (bool): flag to only run once and stop
        debug (bool): debug flag to propagate exceptions
    """
    start_time = time.time()

    try:
        logger.info('starting materialization request')
        if only_dataset:
            ret = await rest_client.request('POST', f'/request/{only_dataset}', {'num': num})
        else:
            ret = await rest_client.request('POST', '/', {'num': num})
        materialization_id = ret['result']
        logger.info(f'waiting for materialization request {materialization_id}')

        while True:
            await asyncio.sleep(10)
            ret = await rest_client.request('GET', f'/status/{materialization_id}')
            if ret['status'] == 'complete':
                logger.info(f'materialization request {materialization_id} complete')
                break
            elif ret['status'] == 'error':
                logger.warning(f'materialization request {materialization_id} failed')
                break
    except Exception as e:
        logger.warning('materialization error', exc_info=True)
        if debug or run_once:
            raise

    if not run_once:
        # run again after 10 minute delay
        stop_time = time.time()
        delay = max(60*10 - (stop_time-start_time), 60*5)
        IOLoop.current().call_later(delay, run, rest_client, only_dataset, num, run_once, debug)


def main():
    import argparse
    parser = argparse.ArgumentParser(description='run a scheduled task once')
    parser.add_argument('-t', '--token', default=os.environ.get('ICEPROD_TOKEN', None), help='auth token')
    parser.add_argument('-d', '--dataset', type=str, default=None, help='dataset id (optional)')
    parser.add_argument('-n', '--num', type=int, default=100, help='number of jobs per dataset to buffer')
    parser.add_argument('--log-level', default='info', help='log level')
    parser.add_argument('--debug', default=False, action='store_true', help='debug enabled')

    args = parser.parse_args()
    args = vars(args)

    logformat='%(asctime)s %(levelname)s %(name)s %(module)s:%(lineno)s - %(message)s'
    logging.basicConfig(format=logformat, level=getattr(logging, args['log_level'].upper()))

    from rest_tools.client import RestClient
    rpc = RestClient('https://materialization.iceprod.icecube.aq', args['token'])

    import asyncio
    asyncio.run(run(rpc, only_dataset=args['dataset'], num=args['num'], run_once=True, debug=args['debug']))

if __name__ == '__main__':
    main()



import logging
import json
import uuid
import math
from collections import defaultdict

import tornado.web
import pymongo
import motor

from iceprod.core import dataclasses
from iceprod.core.resources import Resources
from iceprod.server.rest import RESTHandler, RESTHandlerSetup, authorization
from iceprod.server.util import nowstr, task_statuses, task_status_sort

logger = logging.getLogger('rest.tasks')

def setup(config, *args, **kwargs):
    """
    Setup method for Tasks REST API.

    Sets up any database connections or other prerequisites.

    Args:
        config (dict): an instance of :py:class:`iceprod.server.config`.

    Returns:
        list: Routes for logs, which can be passed to :py:class:`tornado.web.Application`.
    """
    cfg_rest = config.get('rest',{}).get('tasks',{})
    db_cfg = cfg_rest.get('database',{})

    # add indexes
    db = pymongo.MongoClient(**db_cfg).tasks
    if 'task_id_index' not in db.tasks.index_information():
        db.tasks.create_index('task_id', name='task_id_index', unique=True)
    if 'dataset_id_index' not in db.tasks.index_information():
        db.tasks.create_index('dataset_id', name='dataset_id_index', unique=False)
    if 'job_id_index' not in db.tasks.index_information():
        db.tasks.create_index('job_id', name='job_id_index', unique=False)
    if 'status_index' not in db.tasks.index_information():
        db.tasks.create_index('status', name='status_index', unique=False)
    if 'priority_index' not in db.tasks.index_information():
        db.tasks.create_index([('status',pymongo.ASCENDING),('priority',pymongo.DESCENDING)], name='priority_index', unique=False)

    if 'dataset_id_index' not in db.dataset_files.index_information():
        db.dataset_files.create_index('dataset_id', name='dataset_id_index', unique=False)
    if 'task_id_index' not in db.dataset_files.index_information():
        db.dataset_files.create_index('task_id', name='task_id_index', unique=False)

    handler_cfg = RESTHandlerSetup(config, *args, **kwargs)
    handler_cfg.update({
        'database': motor.motor_tornado.MotorClient(**db_cfg).tasks,
    })

    return [
        (r'/tasks', MultiTasksHandler, handler_cfg),
        (r'/tasks/(?P<task_id>\w+)', TasksHandler, handler_cfg),
        (r'/tasks/(?P<task_id>\w+)/status', TasksStatusHandler, handler_cfg),
        (r'/task_actions/queue', TasksActionsQueueHandler, handler_cfg),
        (r'/task_actions/bulk_status/(?P<status>\w+)', TaskBulkStatusHandler, handler_cfg),
        (r'/task_actions/process', TasksActionsProcessingHandler, handler_cfg),
        (r'/task_counts/status', TaskCountsStatusHandler, handler_cfg),
        (r'/tasks/(?P<task_id>\w+)/task_actions/reset', TasksActionsErrorHandler, handler_cfg),
        (r'/tasks/(?P<task_id>\w+)/task_actions/complete', TasksActionsCompleteHandler, handler_cfg),
        (r'/datasets/(?P<dataset_id>\w+)/tasks', DatasetMultiTasksHandler, handler_cfg),
        (r'/datasets/(?P<dataset_id>\w+)/tasks/(?P<task_id>\w+)', DatasetTasksHandler, handler_cfg),
        (r'/datasets/(?P<dataset_id>\w+)/tasks/(?P<task_id>\w+)/status', DatasetTasksStatusHandler, handler_cfg),
        (r'/datasets/(?P<dataset_id>\w+)/task_summaries/status', DatasetTaskSummaryStatusHandler, handler_cfg),
        (r'/datasets/(?P<dataset_id>\w+)/task_counts/status', DatasetTaskCountsStatusHandler, handler_cfg),
        (r'/datasets/(?P<dataset_id>\w+)/task_counts/name_status', DatasetTaskCountsNameStatusHandler, handler_cfg),
        (r'/datasets/(?P<dataset_id>\w+)/task_actions/bulk_status/(?P<status>\w+)', DatasetTaskBulkStatusHandler, handler_cfg),
        (r'/datasets/(?P<dataset_id>\w+)/task_actions/bulk_requirements/(?P<name>\w+)', DatasetTaskBulkRequirementsHandler, handler_cfg),
        (r'/datasets/(?P<dataset_id>\w+)/task_stats', DatasetTaskStatsHandler, handler_cfg),
        (r'/datasets/(?P<dataset_id>\w+)/files', DatasetMultiFilesHandler, handler_cfg),
        (r'/datasets/(?P<dataset_id>\w+)/files/(?P<task_id>\w+)', DatasetTaskFilesHandler, handler_cfg),
    ]

class BaseHandler(RESTHandler):
    def initialize(self, database=None, **kwargs):
        super(BaseHandler, self).initialize(**kwargs)
        self.db = database

class MultiTasksHandler(BaseHandler):
    """
    Handle multi tasks requests.
    """
    @authorization(roles=['admin','system','client'])
    async def get(self):
        """
        Get task entries.

        Params (optional):
            status: | separated list of task status to filter by
            keys: | separated list of keys to return for each task
            sort: | separated list of sort key=values, with values of 1 or -1
            limit: number of tasks to return

        Returns:
            dict: {'tasks': [<task>]}
        """
        filters = {}

        status = self.get_argument('status', None)
        if status:
            filters['status'] = {'$in': status.split('|')}

        sort = self.get_argument('sort', None)
        mongo_sort = []
        if sort:
            for s in sort.split('|'):
                if '=' in s:
                    name, order = s.split('=', 1)
                    if order == '-1':
                        mongo_sort.append((name, pymongo.DESCENDING))
                    else:
                        mongo_sort.append((name, pymongo.ASCENDING))
                else:
                    mongo_sort.append((s, pymongo.ASCENDING))

        limit = self.get_argument('limit', 0)
        if limit:
            try:
                limit = int(limit)
            except Exception:
                limit = 0

        projection = {x:True for x in self.get_argument('keys','').split('|') if x}
        projection['_id'] = False

        ret = []
        async for row in self.db.tasks.find(filters, projection=projection, sort=mongo_sort, limit=limit):
            ret.append(row)
        self.write({'tasks': ret})

    @authorization(roles=['admin','system','client'])
    async def post(self):
        """
        Create a task entry.

        Body should contain the task data.

        Returns:
            dict: {'result': <task_id>}
        """
        data = json.loads(self.request.body)

        # validate first
        req_fields = {
            'dataset_id': str,
            'job_id': str,
            'task_index': int,
            'job_index': int,
            'name': str,
            'depends': list,
            'requirements': dict,
        }
        for k in req_fields:
            if k not in data:
                raise tornado.web.HTTPError(400, reason='missing key: '+k)
            if not isinstance(data[k], req_fields[k]):
                r = 'key {} should be of type {}'.format(k, req_fields[k])
                raise tornado.web.HTTPError(400, reason=r)

        # set some fields
        task_id = uuid.uuid1().hex
        data.update({
            'task_id': task_id,
            'status_changed': nowstr(),
            'failures': 0,
            'evictions': 0,
            'walltime': 0.0,
            'walltime_err': 0.0,
            'walltime_err_n': 0,
            'site': '',
        })
        if 'status' not in data:
            data['status'] = 'waiting'
        if 'priority' not in data:
            data['priority'] = 1.

        ret = await self.db.tasks.insert_one(data)
        self.set_status(201)
        self.write({'result': task_id})
        self.finish()

class TasksHandler(BaseHandler):
    """
    Handle single task requests.
    """
    @authorization(roles=['admin','client','system','pilot'])
    async def get(self, task_id):
        """
        Get a task entry.

        Args:
            task_id (str): the task id

        Returns:
            dict: task entry
        """
        ret = await self.db.tasks.find_one({'task_id':task_id},
                projection={'_id':False})
        if not ret:
            self.send_error(404, reason="Task not found")
        else:
            self.write(ret)
            self.finish()

    @authorization(roles=['admin','client','system','pilot'])
    async def patch(self, task_id):
        """
        Update a task entry.

        Body should contain the task data to update.  Note that this will
        perform a merge (not replace).

        Args:
            task_id (str): the task id

        Returns:
            dict: updated task entry
        """
        data = json.loads(self.request.body)
        if not data:
            raise tornado.web.HTTPError(400, reason='Missing update data')

        ret = await self.db.tasks.find_one_and_update({'task_id':task_id},
                {'$set':data},
                projection={'_id':False},
                return_document=pymongo.ReturnDocument.AFTER)
        if not ret:
            self.send_error(404, reason="Task not found")
        else:
            self.write(ret)
            self.finish()

class TasksStatusHandler(BaseHandler):
    """
    Handle single task requests.
    """
    @authorization(roles=['admin','client','system', 'pilot'])
    async def put(self, task_id):
        """
        Set a task status.

        Body should have {'status': <new_status>}

        Args:
            task_id (str): the task id

        Returns:
            dict: empty dict
        """
        data = json.loads(self.request.body)
        if (not data) or 'status' not in data:
            raise tornado.web.HTTPError(400, reason='Missing status in body')
        if data['status'] not in task_statuses:
            raise tornado.web.HTTPError(400, reason='Bad status')
        update_data = {
            'status': data['status'],
            'status_changed': nowstr(),
        }

        ret = await self.db.tasks.update_one({'task_id':task_id},
                {'$set':update_data})
        if (not ret) or ret.modified_count < 1:
            self.send_error(404, reason="Task not found")
        else:
            self.write({})
            self.finish()

class TaskCountsStatusHandler(BaseHandler):
    """
    Handle task summary grouping by status.
    """
    @authorization(roles=['admin','client','system'])
    async def get(self):
        """
        Get the task counts for all tasks, group by status.

        Returns:
            dict: {<status>: num}
        """
        ret = {}
        for status in task_statuses:
            ret[status] = await self.db.tasks.count_documents({"status":status})

        ret2 = {}
        for k in sorted(ret, key=task_status_sort):
            ret2[k] = ret[k]
        self.write(ret2)
        self.finish()

class DatasetMultiTasksHandler(BaseHandler):
    """
    Handle multi tasks requests.
    """
    @authorization(roles=['admin','client','system'], attrs=['dataset_id:read'])
    async def get(self, dataset_id):
        """
        Get task entries.

        Params (optional):
            status: | separated list of task status to filter by
            job_id: job_id to filter by
            job_index: job_index to filter by
            keys: | separated list of keys to return for each task

        Args:
            dataset_id (str): dataset id

        Returns:
            dict: {'task_id': {task data}}
        """
        filters = {'dataset_id':dataset_id}

        status = self.get_argument('status', None)
        if status:
            filters['status'] = {'$in': status.split('|')}

        job_id = self.get_argument('job_id', None)
        if job_id:
            filters['job_id'] = job_id

        job_index = self.get_argument('job_index', None)
        if job_index:
            try:
                filters['job_index'] = int(job_index)
            except ValueError:
                raise tornado.web.HTTPError(400, reason='Bad argument "job_index": must be integer')

        projection = {'_id': False}
        keys = self.get_argument('keys','')
        if keys:
            projection.update({x:True for x in keys.split('|') if x})
            projection['task_id'] = True

        ret = {}
        async for row in self.db.tasks.find(filters, projection=projection):
            ret[row['task_id']] = row
        self.write(ret)

class DatasetTasksHandler(BaseHandler):
    """
    Handle single task requests.
    """
    @authorization(roles=['admin','client','system'], attrs=['dataset_id:read'])
    async def get(self, dataset_id, task_id):
        """
        Get a task entry.

        Args:
            dataset_id (str): dataset id
            task_id (str): the task id

        Params (optional):
            keys: | separated list of keys to return for each task

        Returns:
            dict: task entry
        """
        projection = {'_id': False}
        keys = self.get_argument('keys','')
        if keys:
            projection.update({x:True for x in keys.split('|') if x})
            projection['task_id'] = True

        ret = await self.db.tasks.find_one({'task_id':task_id,'dataset_id':dataset_id},
                projection=projection)
        if not ret:
            self.send_error(404, reason="Task not found")
        else:
            self.write(ret)
            self.finish()

class DatasetTasksStatusHandler(BaseHandler):
    """
    Handle single task requests.
    """
    @authorization(roles=['admin','client','system'], attrs=['dataset_id:write'])
    async def put(self, dataset_id, task_id):
        """
        Set a task status.

        Body should have {'status': <new_status>}

        Args:
            dataset_id (str): dataset id
            task_id (str): the task id

        Returns:
            dict: empty dict
        """
        data = json.loads(self.request.body)
        if (not data) or 'status' not in data:
            raise tornado.web.HTTPError(400, reason='Missing status in body')
        if data['status'] not in task_statuses:
            raise tornado.web.HTTPError(400, reason='Bad status')
        update_data = {
            'status': data['status'],
            'status_changed': nowstr(),
        }
        if data['status'] == 'reset':
            update_data['failures'] = 0

        ret = await self.db.tasks.update_one({'task_id':task_id,'dataset_id':dataset_id},
                {'$set':update_data})
        if (not ret) or ret.modified_count < 1:
            self.send_error(404, reason="Task not found")
        else:
            self.write({})
            self.finish()

class DatasetTaskSummaryStatusHandler(BaseHandler):
    """
    Handle task summary grouping by status.
    """
    @authorization(roles=['admin','client','system'], attrs=['dataset_id:read'])
    async def get(self, dataset_id):
        """
        Get the task summary for all tasks in a dataset, group by status.

        Args:
            dataset_id (str): dataset id

        Returns:
            dict: {<status>: [<task_id>,]}
        """
        cursor = self.db.tasks.find({'dataset_id':dataset_id},
                projection={'_id':False,'status':True,'task_id':True})
        ret = defaultdict(list)
        async for row in cursor:
            ret[row['status']].append(row['task_id'])
        ret2 = {}
        for k in sorted(ret, key=task_status_sort):
            ret2[k] = ret[k]
        self.write(ret2)
        self.finish()

class DatasetTaskCountsStatusHandler(BaseHandler):
    """
    Handle task summary grouping by status.
    """
    @authorization(roles=['admin','client','system'], attrs=['dataset_id:read'])
    async def get(self, dataset_id):
        """
        Get the task counts for all tasks in a dataset, group by status.

        Args:
            dataset_id (str): dataset id

        Returns:
            dict: {<status>: num}
        """
        cursor = self.db.tasks.aggregate([
            {'$match':{'dataset_id':dataset_id}},
            {'$group':{'_id':'$status', 'total': {'$sum':1}}},
        ])
        ret = {}
        async for row in cursor:
            ret[row['_id']] = row['total']
        ret2 = {}
        for k in sorted(ret, key=task_status_sort):
            ret2[k] = ret[k]
        self.write(ret2)
        self.finish()

class DatasetTaskCountsNameStatusHandler(BaseHandler):
    """
    Handle task summary grouping by name and status.
    """
    @authorization(roles=['admin','client','system'], attrs=['dataset_id:read'])
    async def get(self, dataset_id):
        """
        Get the task counts for all tasks in a dataset, group by name,status.

        Args:
            dataset_id (str): dataset id

        Returns:
            dict: {<name>: {<status>: num}}
        """
        cursor = self.db.tasks.aggregate([
            {'$match':{'dataset_id':dataset_id}},
            {'$group':{
                '_id':{'name':'$name','status':'$status'},
                'ordering':{'$first':'$task_index'},
                'total': {'$sum':1}
            }},
        ])
        ret = defaultdict(dict)
        ordering = {}
        async for row in cursor:
            ret[row['_id']['name']][row['_id']['status']] = row['total']
            ordering[row['_id']['name']] = row['ordering']
        ret2 = {}
        for k in sorted(ordering, key=lambda n:ordering[n]):
            ret2[k] = ret[k]
        self.write(ret2)
        self.finish()

class DatasetTaskStatsHandler(BaseHandler):
    """
    Handle task stats
    """
    @authorization(roles=['admin','client','system'], attrs=['dataset_id:read'])
    async def get(self, dataset_id):
        """
        Get the task statistics for all tasks in a dataset, group by name.

        Args:
            dataset_id (str): dataset id

        Returns:
            dict: {<name>: {<stat>: <value>}}
        """
        cursor = self.db.tasks.aggregate([
            {'$match':{'dataset_id':dataset_id, 'status':'complete'}},
            {'$group':{'_id':'$name',
                'count': {'$sum': 1},
                'gpu': {'$sum': '$requirements.gpu'},
                'total_hrs': {'$sum': '$walltime'},
                'total_err_hrs': {'$sum': '$walltime_err'},
                'avg_hrs': {'$avg': '$walltime'},
                'stddev_hrs': {'$stdDevSamp': '$walltime'},
                'min_hrs': {'$min': '$walltime'},
                'max_hrs': {'$max': '$walltime'},
                'ordering': {'$first': '$task_index'},
            }},
        ])
        ret = {}
        ordering = {}
        async for row in cursor:
            denom = row['total_hrs'] + row['total_err_hrs']
            row['efficiency'] = row['total_hrs']/denom if denom > 0 else 0.0
            name = row.pop('_id')
            ordering[name] = row.pop('ordering')
            ret[name] = row
        ret2 = {}
        for k in sorted(ordering, key=lambda n:ordering[n]):
            ret2[k] = ret[k]
        self.write(ret2)
        self.finish()

class TasksActionsQueueHandler(BaseHandler):
    """
    Handle task action for waiting -> queued.
    """
    @authorization(roles=['admin','client'])
    async def post(self):
        """
        Take a number of waiting tasks and queue them.

        Order by priority.

        Body args (json):
            num_tasks: int

        Returns:
            dict: {queued: num tasks queued}
        """
        data = json.loads(self.request.body)
        num_tasks = data.get('num_tasks', 100)

        query = {'status': 'waiting'}
        val = {'$set': {'status': 'queued'}}

        queued = 0
        while queued < num_tasks:
            ret = await self.db.tasks.find_one_and_update(query, val,
                    projection={'_id':False,'task_id':True},
                    sort=[('priority', -1)])
            if not ret:
                logger.debug('no more tasks to queue')
                break
            queued += 1
        logger.info(f'queued {queued} tasks')
        self.write({'queued': queued})

class TasksActionsProcessingHandler(BaseHandler):
    """
    Handle task action for queued -> processing.
    """
    @authorization(roles=['admin','client','pilot'])
    async def post(self):
        """
        Take one queued task, set its status to processing, and return it.

        Body args (json):
            requirements: dict
            query_params: (optional) dict of mongodb params

        Returns:
            dict: <task dict>
        """
        filter_query = {'status':'queued'}
        sort_by = [('priority',-1)]
        site = 'unknown'
        if self.request.body:
            data = json.loads(self.request.body)
            # handle requirements
            reqs = data.get('requirements', {})
            req_filters = []
            for k in reqs:
                if k == 'gpu' and reqs[k] > 0:
                    val = {'$lte': reqs[k], '$gte': 1}
                    req_filters.append({'requirements.'+k: val})
                    continue
                elif isinstance(reqs[k], (int,float)):
                    val = {'$lte': reqs[k]}
                else:
                    val = reqs[k]
                req_filters.append({'$or': [
                    {'requirements.'+k: {'$exists': False}},
                    {'requirements.'+k: val},
                ]})
            if req_filters:
                filter_query['$and'] = req_filters
            if 'site' in reqs:
                site = reqs['site']
            # handle query_params
            params = data.get('query_params', {})
            for k in params:
                if k in filter_query:
                    raise tornado.web.HTTPError(400, reason=f'param {k} would override an already set filter')
                filter_query[k] = params[k]
        print('filter_query', filter_query)
        ret = await self.db.tasks.find_one_and_update(filter_query,
                {'$set':{'status':'processing'}},
                projection={'_id':False},
                sort=sort_by,
                return_document=pymongo.ReturnDocument.AFTER)
        if not ret:
            logger.info('filter_query: %r', filter_query)
            self.send_error(404, reason="Task not found")
        else:
            self.module.statsd.incr('site.{}.task_processing'.format(site))
            self.write(ret)
            self.finish()

class TasksActionsErrorHandler(BaseHandler):
    """
    Handle task action on error (* -> reset).
    """
    @authorization(roles=['admin','client','pilot'])
    async def post(self, task_id):
        """
        Take one task, set its status to reset.

        Args:
            task_id (str): task id

        Body args (json):
            time_used (int): (optional) time used to run task, in seconds
            resources (dict): (optional) resources used by task
            site (str): (optional) site the task was running at
            reason (str): (optional) reason for error

        Returns:
            dict: {}  empty dict
        """
        filter_query = {'task_id': task_id, 'status': {'$ne': 'complete'}}
        update_query = defaultdict(dict,{
            '$set': {
                'status': 'reset',
                'status_changed': nowstr(),
            },
            '$inc': {
                'failures': 1,
            },
        })
        if self.request.body:
            data = json.loads(self.request.body)
            task = await self.db.tasks.find_one(filter_query)
            if 'time_used' in data:
                update_query['$inc']['walltime_err_n'] = 1
                update_query['$inc']['walltime_err'] = data['time_used']/3600.
            elif 'resources' in data and 'time' in data['resources']:
                update_query['$inc']['walltime_err_n'] = 1
                update_query['$inc']['walltime_err'] = data['resources']['time']
            for k in ('cpu','memory','disk','time'):
                if 'resources' in data and k in data['resources']:
                    try:
                        new_val = float(data['resources'][k])
                        old_val = task['requirements'][k] if k in task['requirements'] else Resources.defaults[k]
                        if k == 'cpu': # special handling for cpu
                            if new_val <= 1.1 or new_val > 20:
                                continue
                            if new_val < old_val*1.1:
                                continue
                            new_val = old_val+1 # increase linearly
                        else:
                            new_val *= 1.5 # increase new request by 1.5
                        if isinstance(Resources.defaults[k], (int, list)):
                            new_val = math.ceil(new_val)
                    except Exception:
                        logger.info('error converting requirement %r',
                                    data['resources'][k], exc_info=True)
                    else:
                        update_query['$max']['requirements.'+k] = new_val
            site = 'unknown'
            if 'site' in data:
                site = data['site']
                update_query['$set']['site'] = site
            if self.module and self.module.statsd and 'reason' in data and data['reason']:
                reason = 'other'
                reasons = [
                    ('Exception: failed to download', 'download_failure'),
                    ('Exception: failed to upload', 'upload_failure'),
                    ('Exception: module failed', 'module_failure'),
                    ('Resource overusage for cpu', 'cpu_overuse'),
                    ('Resource overusage for gpu', 'gpu_overuse'),
                    ('Resource overusage for memory', 'memory_overuse'),
                    ('Resource overusage for disk', 'disk_overuse'),
                    ('Resource overusage for time', 'time_overuse'),
                    ('pilot SIGTERM', 'sigterm'),
                    ('killed', 'killed'),
                ]
                for text,r in reasons:
                    if text in data['reason']:
                        reason = r
                        break
                self.module.statsd.incr('site.{}.task_reset.{}'.format(site, reason))
        ret = await self.db.tasks.find_one_and_update(filter_query,
                update_query,
                projection={'_id':False})
        if not ret:
            logger.info('filter_query: %r', filter_query)
            self.send_error(404, reason="Task not found")
        else:
            self.write(ret)
            self.finish()

class TasksActionsCompleteHandler(BaseHandler):
    """
    Handle task action on processing -> complete.
    """
    @authorization(roles=['admin','client','pilot'])
    async def post(self, task_id):
        """
        Take one task, set its status to complete.

        Args:
            task_id (str): task id

        Body args (json):
            time_used (int): (optional) time used to run task, in seconds
            site (str): (optional) site the task was running at

        Returns:
            dict: {}  empty dict
        """
        filter_query = {'task_id': task_id, 'status': 'processing'}
        update_query = {
            '$set': {
                'status': 'complete',
                'status_changed': nowstr(),
            },
        }
        if self.request.body:
            data = json.loads(self.request.body)
            if 'time_used' in data:
                update_query['$set']['walltime'] = data['time_used']/3600.
            site = 'unknown'
            if 'site' in data:
                site = data['site']
                update_query['$set']['site'] = site
            self.module.statsd.incr('site.{}.task_complete'.format(site))
        ret = await self.db.tasks.find_one_and_update(filter_query,
                update_query,
                projection={'_id':False})
        if not ret:
            logger.info('filter_query: %r', filter_query)
            self.send_error(404, reason="Task not found or not processing")
        else:
            self.write(ret)
            self.finish()

class TaskBulkStatusHandler(BaseHandler):
    """
    Update the status of multiple tasks at once.
    """
    @authorization(roles=['admin','client','system'])
    async def post(self, status):
        """
        Set multiple tasks' status.

        Body should have {'tasks': [<task_id>, <task_id>, ...]}

        Args:
            status (str): the status

        Returns:
            dict: empty dict
        """
        data = json.loads(self.request.body)
        if (not data) or 'tasks' not in data or not data['tasks']:
            raise tornado.web.HTTPError(400, reason='Missing tasks in body')
        tasks = list(data['tasks'])
        if len(tasks) > 100000:
            raise tornado.web.HTTPError(400, reason='Too many tasks specified (limit: 100k)')
        if status not in task_statuses:
            raise tornado.web.HTTPError(400, reason='Bad status')
        query = {
            'task_id': {'$in': tasks},
        }
        update_data = {
            'status': status,
            'status_changed': nowstr(),
        }
        if status == 'reset':
            update_data['failures'] = 0

        ret = await self.db.tasks.update_many(query, {'$set':update_data})
        if (not ret) or ret.modified_count < 1:
            self.send_error(404, reason="Tasks not found")
        else:
            self.write({})
            self.finish()

class DatasetTaskBulkStatusHandler(BaseHandler):
    """
    Update the status of multiple tasks at once.
    """
    @authorization(roles=['admin','client','system'], attrs=['dataset_id:write'])
    async def post(self, dataset_id, status):
        """
        Set multiple tasks' status.

        Body should have {'tasks': [<task_id>, <task_id>, ...]}

        Args:
            dataset_id (str): dataset id
            status (str): the status

        Returns:
            dict: empty dict
        """
        data = json.loads(self.request.body)
        if (not data) or 'tasks' not in data or not data['tasks']:
            raise tornado.web.HTTPError(400, reason='Missing tasks in body')
        tasks = list(data['tasks'])
        if len(tasks) > 100000:
            raise tornado.web.HTTPError(400, reason='Too many tasks specified (limit: 100k)')
        if status not in task_statuses:
            raise tornado.web.HTTPError(400, reason='Bad status')
        query = {
            'dataset_id': dataset_id,
            'task_id': {'$in': tasks},
        }
        update_data = {
            'status': status,
            'status_changed': nowstr(),
        }
        if status == 'reset':
            update_data['failures'] = 0

        ret = await self.db.tasks.update_many(query, {'$set':update_data})
        if (not ret) or ret.modified_count < 1:
            self.send_error(404, reason="Tasks not found")
        else:
            self.write({})
            self.finish()

class DatasetTaskBulkRequirementsHandler(BaseHandler):
    """
    Update the requirements of multiple tasks at once.
    """
    @authorization(roles=['admin','client','system'], attrs=['dataset_id:write'])
    async def patch(self, dataset_id, name):
        """
        Set multiple tasks' requirements. Sets for all tasks in a dataset
        with the specified name.

        Body should have {<resource>: <requirement>}.

        Args:
            dataset_id (str): dataset id
            name (str): the task name

        Returns:
            dict: empty dict
        """
        valid_req_keys = set(Resources.defaults)
        valid_req_keys.add('os')
        valid_req_keys.add('site')
        
        data = json.loads(self.request.body)
        if (not data):
            raise tornado.web.HTTPError(400, reason='Missing body')
        elif set(data) - valid_req_keys:
            raise tornado.web.HTTPError(400, reason='Invalid resource types')

        reqs = {}
        for key in valid_req_keys.intersection(data):
            val = data[key]
            if key == 'os':
                if not isinstance(val, list):
                    raise tornado.web.HTTPError(400, reason='Bad type for {}, should be list'.format(key))
            elif key in Resources.defaults and isinstance(Resources.defaults[key], (int, list)):
                if not isinstance(val, int):
                    raise tornado.web.HTTPError(400, reason='Bad type for {}, should be int'.format(key))
            elif key in Resources.defaults and isinstance(Resources.defaults[key], float):
                if not isinstance(val, (int,float)):
                    raise tornado.web.HTTPError(400, reason='Bad type for {}, should be float'.format(key))
            else:
                val = str(val)
            reqs['requirements.'+key] = val

        query = {
            'dataset_id': dataset_id,
            'name': name,
        }
        ret = await self.db.tasks.update_many(query,
                {'$max':reqs})
        if (not ret) or ret.modified_count < 1:
            self.send_error(404, reason="Tasks not found")
        else:
            self.write({})
            self.finish()

class DatasetMultiFilesHandler(BaseHandler):
    """
    Handle multi files requests, by dataset.
    """
    @authorization(roles=['admin','system','client'], attrs=['dataset_id:read'])
    async def get(self, dataset_id):
        """
        Get dataset_files entries.

        Args:
            dataset_id (str): dataset id

        Returns:
            dict: {'files': [<file>]}
        """
        filters = {'dataset_id': dataset_id}
        projection = {'_id':False, 'dataset_id':False, 'task_id':False}
        ret = []
        async for row in self.db.dataset_files.find(filters, projection=projection):
            ret.append(row)
        self.write({'files': ret})

    @authorization(roles=['admin','system','client'], attrs=['dataset_id:write'])
    async def post(self, dataset_id):
        """
        Create a dataset_files entry.

        Body should contain the file data.

        Parameters:
            filename (str): the full url filename
            movement (str): [input | output | both]
            job_index (int): the job index to add to
            task_name (str): the name of the task
            local (str): (optional) the local filename the task sees
            transfer (str): whether to transfer the file (can be bool or str)
            compression (str): whether to automatically compress/decompress the file

        Returns:
            dict: {'result': <task_id>}
        """
        data = json.loads(self.request.body)

        # validate first
        req_fields = {
            'filename': str,
            'movement': str,
            'job_index': int,
            'task_name': str,
        }
        for k in req_fields:
            if k not in data:
                raise tornado.web.HTTPError(400, reason='missing key: '+k)
            if not isinstance(data[k], req_fields[k]):
                r = 'key {} should be of type {}'.format(k, req_fields[k])
                raise tornado.web.HTTPError(400, reason=r)

        # find the task referred to
        filters = {
            'dataset_id': dataset_id,
            'job_index': data['job_index'],
            'name': data['task_name'],
        }
        ret = await self.db.tasks.find_one(filters)
        if not ret:
            raise tornado.web.HTTPError(400, reason='task referred to not found')

        # set some fields
        file_data = dataclasses.Data()
        file_data.update({
            'task_id': ret['task_id'],
            'dataset_id': dataset_id,
            'remote': data['filename'],
            'movement': data['movement'],
        })
        if 'local' in data:
            if not isinstance(data['local'], str):
                r = 'key {} should be of type {}'.format('local', str)
                raise tornado.web.HTTPError(400, reason=r)
            file_data['local'] = data['local']
        if 'transfer' in data:
            if not isinstance(data['transfer'], (str,bool)):
                r = 'key {} should be of type {}'.format('transfer', str)
                raise tornado.web.HTTPError(400, reason=r)
            file_data['transfer'] = data['transfer']
        if 'compression' in data:
            if not isinstance(data['compression'], (str,bool)):
                r = 'key {} should be of type {}'.format('compression', str)
                raise tornado.web.HTTPError(400, reason=r)
            file_data['compression'] = data['compression']
        if not file_data.valid():
            raise tornado.web.HTTPError(400, reason='invalid file data')

        ret = await self.db.dataset_files.insert_one(dict(file_data))
        self.set_status(201)
        self.write({'result': file_data['task_id']})
        self.finish()

class DatasetTaskFilesHandler(BaseHandler):
    """
    Handle multi files requests, by task.
    """
    @authorization(roles=['admin','system','client','pilot'], attrs=['dataset_id:read'])
    async def get(self, dataset_id, task_id):
        """
        Get dataset_files entries.

        Args:
            dataset_id (str): dataset id
            task_id (str): task_id

        Returns:
            dict: {'files': [<file>]}
        """
        filters = {'dataset_id': dataset_id, 'task_id': task_id}
        projection = {'_id':False, 'dataset_id':False, 'task_id':False}
        ret = []
        async for row in self.db.dataset_files.find(filters, projection=projection):
            ret.append(row)
        self.write({'files': ret})

    @authorization(roles=['admin','system','client'], attrs=['dataset_id:write'])
    async def post(self, dataset_id, task_id):
        """
        Create a dataset_files entry.

        Body should contain the file data.

        Parameters:
            filename (str): the full url filename
            movement (str): [input | output | both]
            local (str): (optional) the local filename the task sees
            transfer (str): whether to transfer the file (can be bool or str)
            compression (str): whether to automatically compress/decompress the file

        Returns:
            dict: {}
        """
        data = json.loads(self.request.body)

        # validate first
        req_fields = {
            'filename': str,
            'movement': str,
        }
        for k in req_fields:
            if k not in data:
                raise tornado.web.HTTPError(400, reason='missing key: '+k)
            if not isinstance(data[k], req_fields[k]):
                r = 'key {} should be of type {}'.format(k, req_fields[k])
                raise tornado.web.HTTPError(400, reason=r)

        # set some fields
        file_data = dataclasses.Data()
        file_data.update({
            'task_id': task_id,
            'dataset_id': dataset_id,
            'remote': data['filename'],
            'movement': data['movement'],
        })
        if 'local' in data:
            if not isinstance(data['local'], str):
                r = 'key {} should be of type {}'.format('local', str)
                raise tornado.web.HTTPError(400, reason=r)
            file_data['local'] = data['local']
        if 'transfer' in data:
            if not isinstance(data['transfer'], (str,bool)):
                r = 'key {} should be of type {}'.format('transfer', str)
                raise tornado.web.HTTPError(400, reason=r)
            file_data['transfer'] = data['transfer']
        if 'compression' in data:
            if not isinstance(data['compression'], (str,bool)):
                r = 'key {} should be of type {}'.format('compression', str)
                raise tornado.web.HTTPError(400, reason=r)
            file_data['compression'] = data['compression']
        if not file_data.valid():
            raise tornado.web.HTTPError(400, reason='invalid file data')

        ret = await self.db.dataset_files.insert_one(dict(file_data))
        self.set_status(201)
        self.write({})
        self.finish()

    @authorization(roles=['admin','system','client'], attrs=['dataset_id:write'])
    async def delete(self, dataset_id, task_id):
        """
        Delete dataset_files entries.

        Args:
            dataset_id (str): dataset id
            task_id (str): task_id

        Returns:
            dict: {}
        """
        filters = {'dataset_id': dataset_id, 'task_id': task_id}
        await self.db.dataset_files.delete_many(filters)
        self.write({})



"""
Suspend any jobs and tasks from datasets that are suspended / errors.

Initial delay: rand(15 minutes)
Periodic delay: 60 minutes
"""

import logging
import random
import time

from tornado.ioloop import IOLoop

logger = logging.getLogger('dataset_cleanup')

def dataset_cleanup(module):
    """
    Initial entrypoint.

    Args:
        module (:py:class:`iceprod.server.modules.schedule`): schedule module
    """
    # initial delay
    IOLoop.current().call_later(random.randint(10,60*10), run, module.rest_client)

async def run(rest_client, debug=False):
    """
    Actual runtime / loop.

    Args:
        rest_client (:py:class:`iceprod.core.rest_client.Client`): rest client
        debug (bool): debug flag to propagate exceptions
    """
    start_time = time.time()
    datasets = await rest_client.request('GET', '/dataset_summaries/status')
    dataset_ids = []
    if 'suspended' in datasets:
        dataset_ids.extend(list(datasets['suspended']))
    if 'errors' in datasets:
        dataset_ids.extend(list(datasets['errors']))
    for dataset_id in dataset_ids:
        try:
            jobs = await rest_client.request('GET', f'/datasets/{dataset_id}/job_summaries/status')
            if 'processing' in jobs:
                for job_id in jobs['processing']:
                    args = {'status': 'suspended'}
                    await rest_client.request('PUT', f'/datasets/{dataset_id}/jobs/{job_id}/status', args)

            tasks = await rest_client.request('GET', f'/datasets/{dataset_id}/task_summaries/status')
            task_ids = []
            for status in tasks:
                if status not in ('complete','suspended','failed'):
                    task_ids.extend(tasks[status])
            while task_ids:
                tids = task_ids[:10000]
                task_ids = task_ids[10000:]
                args = {'tasks': tids}
                await rest_client.request('POST', f'/datasets/{dataset_id}/task_actions/bulk_status/suspended', args)

        except Exception:
            logger.error('error cleaning a job/task in dataset %s', dataset_id, exc_info=True)
            if debug:
                raise

    # clean up tasks in suspended/errors jobs in processing datasets
    dataset_ids = []
    if 'processing' in datasets:
        dataset_ids.extend(list(datasets['processing']))
    for dataset_id in dataset_ids:
        try:
            jobs = await rest_client.request('GET', f'/datasets/{dataset_id}/job_summaries/status')
            job_ids = set()
            if 'suspended' in jobs:
                job_ids.update(jobs['suspended'])
            if 'errors' in jobs:
                job_ids.update(jobs['errors'])

            args = {
                'keys': 'task_id|job_id|status'
            }
            tasks = await rest_client.request('GET', f'/datasets/{dataset_id}/tasks', args)
            task_ids = []
            for tasks in tasks.values():
                if tasks['status'] not in ('complete','suspended') and tasks['job_id'] in job_ids:
                    task_ids.append(tasks['task_id'])
            while task_ids:
                tids = task_ids[:10000]
                task_ids = task_ids[10000:]
                args = {'tasks': tids}
                await rest_client.request('POST', f'/datasets/{dataset_id}/task_actions/bulk_status/suspended', args)

        except Exception:
            logger.error('error cleaning a task in dataset %s', dataset_id, exc_info=True)
            if debug:
                raise

    # run again after 60 minute delay
    stop_time = time.time()
    delay = max(60*60 - (stop_time-start_time), 60*10)
    IOLoop.current().call_later(delay, run, rest_client)


def main():
    import argparse
    import asyncio
    import os
    parser = argparse.ArgumentParser(description='run a scheduled task once')
    parser.add_argument('-t', '--token', default=os.environ.get('ICEPROD_TOKEN', None), help='auth token')
    parser.add_argument('--log-level', default='info', help='log level')
    parser.add_argument('--debug', default=False, action='store_true', help='debug enabled')

    args = parser.parse_args()
    args = vars(args)

    logformat='%(asctime)s %(levelname)s %(name)s %(module)s:%(lineno)s - %(message)s'
    logging.basicConfig(format=logformat, level=getattr(logging, args['log_level'].upper()))

    from rest_tools.client import RestClient
    rpc = RestClient('https://iceprod2-api.icecube.wisc.edu', args['token'])

    asyncio.run(run(rpc, debug=args['debug']))

if __name__ == '__main__':
    main()



"""
Update pilots when tasks change state via a user.

Initial delay: rand(5 minutes)
Periodic delay: 5 minutes
"""

import logging
import random
import time
import asyncio
from datetime import datetime

from tornado.ioloop import IOLoop
import requests.exceptions

from iceprod.server.util import str2datetime

logger = logging.getLogger('removed_tasks')

def removed_tasks(module):
    """
    Initial entrypoint.

    Args:
        module (:py:class:`iceprod.server.modules.schedule`): schedule module
    """
    # initial delay
    IOLoop.current().call_later(random.randint(60,60*5), run, module.rest_client)

async def run(rest_client, debug=False):
    """
    Actual runtime / loop.

    Args:
        rest_client (:py:class:`iceprod.core.rest_client.Client`): rest client
        debug (bool): debug flag to propagate exceptions
    """
    start_time = time.time()

    try:
        async def test_pilot(pilot):
            new_tasks = []
            for task_id in pilot['tasks']:
                ret = await rest_client.request('GET', f'/tasks/{task_id}')
                if ret['status'] == 'processing':
                    new_tasks.append(task_id)
            if new_tasks != pilot['tasks']:
                args = {'tasks': new_tasks}
                try:
                    await rest_client.request('PATCH', f'/pilots/{pilot["pilot_id"]}', args)
                except requests.exceptions.HTTPError as e:
                    if e.response.status_code != 404:
                        raise

        awaitables = set()
        pilots = await rest_client.request('GET', '/pilots?keys=pilot_id|tasks')
        for p in pilots.values():
            if 'tasks' in p and p['tasks']:
                awaitables.add(asyncio.create_task(test_pilot(p)))
                if len(awaitables) >= 40:
                    done,pending = await asyncio.wait(awaitables, return_when=asyncio.FIRST_COMPLETED)
                    awaitables = pending
                    for fut in done:
                        await fut
        for fut in asyncio.as_completed(awaitables):
            await fut

    except Exception:
        logger.error('error updating pilot for removed tasks', exc_info=True)
        if debug:
            raise

    # run again after 60 minute delay
    stop_time = time.time()
    delay = max(60*5 - (stop_time-start_time), 60)
    IOLoop.current().call_later(delay, run, rest_client)


def main():
    import argparse
    import os
    parser = argparse.ArgumentParser(description='run a scheduled task once')
    parser.add_argument('-t', '--token', default=os.environ.get('ICEPROD_TOKEN', None), help='auth token')
    parser.add_argument('--log-level', default='info', help='log level')
    parser.add_argument('--debug', default=False, action='store_true', help='debug enabled')

    args = parser.parse_args()
    args = vars(args)

    logformat='%(asctime)s %(levelname)s %(name)s %(module)s:%(lineno)s - %(message)s'
    logging.basicConfig(format=logformat, level=getattr(logging, args['log_level'].upper()))

    from rest_tools.client import RestClient
    rpc = RestClient('https://iceprod2-api.icecube.wisc.edu', args['token'])

    asyncio.run(run(rpc, debug=args['debug']))

if __name__ == '__main__':
    main()



import logging
import json
import uuid

import tornado.web
import pymongo
import motor

from iceprod.server.rest import RESTHandler, RESTHandlerSetup, authorization, catch_error

logger = logging.getLogger('rest.config')

def setup(config, *args, **kwargs):
    """
    Setup method for Config REST API.

    Sets up any database connections or other prerequisites.

    Args:
        config (dict): an instance of :py:class:`iceprod.server.config`.

    Returns:
        list: Routes for config, which can be passed to :py:class:`tornado.web.Application`.
    """
    cfg_rest = config.get('rest',{}).get('config',{})
    db_cfg = cfg_rest.get('database',{})

    # add indexes
    db = pymongo.MongoClient(**db_cfg).config
    if 'dataset_id_index' not in db.config.index_information():
        db.config.create_index('dataset_id', name='dataset_id_index', unique=True)

    handler_cfg = RESTHandlerSetup(config, *args, **kwargs)
    handler_cfg.update({
        'database': motor.motor_tornado.MotorClient(**db_cfg).config,
    })

    return [
        (r'/config/(?P<dataset_id>\w+)', ConfigHandler, handler_cfg),
    ]

class ConfigHandler(RESTHandler):
    """
    Handle config requests.
    """
    def initialize(self, database=None, **kwargs):
        super(ConfigHandler, self).initialize(**kwargs)
        self.db = database

    @authorization(roles=['admin','client','system','pilot'], attrs=['dataset_id:read'])
    async def get(self, dataset_id):
        """
        Get a config.

        Args:
            dataset_id (str): the dataset id of the config

        Returns:
            dict: config
        """
        ret = await self.db.config.find_one({'dataset_id':dataset_id},
                projection={'_id':False, 'dataset_id':False})
        if not ret:
            self.send_error(404, reason="Config not found")
        else:
            self.write(ret)
            self.finish()

    @authorization(roles=['admin'], attrs=['dataset_id:write'])
    async def put(self, dataset_id):
        """
        Set a config.

        Body should contain the config.

        Args:
            dataset_id (str): the dataset id of the config

        Returns:
            dict: empty dict
        """
        data = json.loads(self.request.body)
        if 'dataset_id' not in data:
            data['dataset_id'] = dataset_id
        elif data['dataset_id'] != dataset_id:
            raise tornado.web.HTTPError(400, reason='dataset_id mismatch')
        ret = await self.db.config.replace_one({'dataset_id':dataset_id},
                data, upsert=True)
        self.write({})
        self.finish()



"""
Common functions
"""

from __future__ import absolute_import, division, print_function

import sys
import os
import re
import shutil
import time
import logging
import socket
import subprocess
import tarfile
import urllib
import tempfile
import hashlib
from functools import partial
from contextlib import contextmanager
import asyncio

try:
    import cPickle as pickle
except ImportError:
    import pickle

try:
    import psutil
except ImportError:
    psutil = None

import requests
from requests.packages.urllib3.util.retry import Retry
from requests.adapters import HTTPAdapter
from requests_toolbelt.multipart.encoder import MultipartEncoder
from rest_tools.client import Session, AsyncSession

from iceprod.core import util
from iceprod.core.gridftp import GridFTP
from iceprod.core.jsonUtil import json_encode, json_decode


### Compression Functions ###
_compress_suffixes = ('.tgz','.gz','.tbz2','.tbz','.bz2','.bz',
                     '.lzma2','.lzma','.lz','.xz')
_tar_suffixes = ('.tar', '.tar.gz', '.tgz', '.tar.bz2', '.tbz2', '.tbz',
                '.tar.lzma', '.tar.xz', '.tlz', '.txz')

def uncompress(infile, out_dir=None):
    """Uncompress a file, if possible"""
    files = []
    cur_dir = os.getcwd()
    try:
        if out_dir:
            os.chdir(out_dir)
        logging.info('uncompressing %s',infile)
        if istarred(infile):
            # handle tarfile
            output = subprocess.check_output(['tar','-atf',infile]).decode('utf-8')
            files = [x for x in output.split('\n') if x.strip() and x[-1] != '/']
            if not files:
                raise Exception('no files inside tarfile')
            for f in files:
                if os.path.exists(f):
                    break
            else:
                subprocess.call(['tar','-axf',infile])
        else:
            if infile.endswith('.gz'):
                cmd = 'gzip'
            elif any(infile.endswith(s) for s in ('.bz','.bz2')):
                cmd = 'bzip2'
            elif any(infile.endswith(s) for s in ('.xz','.lzma')):
                cmd = 'xz'
            else:
                logging.info('unknown format: %s',infile)
                raise Exception('unknown format')
            subprocess.call([cmd,'-kdf',infile])
            files.append(infile.rsplit('.',1)[0])
    finally:
        os.chdir(cur_dir)

    logging.info('files: %r', files)
    if len(files) == 1:
        return files[0]
    else:
        return files

def compress(infile,compression='lzma'):
    """Compress a file or directory.
       The compression argument is used as the new file extension"""
    if not istarred('.'+compression) and os.path.isdir(infile):
        outfile = infile+'.tar.'+compression
    else:
        outfile = infile+'.'+compression
    if istarred(outfile):
        dirname, filename = os.path.split(infile)
        subprocess.call(['tar','-acf',outfile,'-C',dirname,filename])
    else:
        if outfile.endswith('.gz'):
            cmd = ['gzip']
        elif any(outfile.endswith(s) for s in ('.bz','.bz2')):
            cmd = ['bzip2']
        elif outfile.endswith('.xz'):
            cmd = ['xz']
        elif outfile.endswith('.lzma'):
            cmd = ['xz','-F','lzma']
        else:
            logging.info('unknown format: %s',infile)
            raise Exception('unknown format')
        subprocess.call(cmd+['-kf',infile])
    return outfile

def iscompressed(infile):
    """Check if a file is a compressed file, based on file name"""
    return any(infile.endswith(s) for s in _compress_suffixes)

def istarred(infile):
    """Check if a file is a tarred file, based on file name"""
    return any(infile.endswith(s) for s in _tar_suffixes)

def cksm(filename,type,buffersize=16384,file=True):
    """Return checksum of file using algorithm specified"""
    if type not in ('md5','sha1','sha256','sha512'):
        raise Exception('cannot get checksum for type %r',type)

    try:
        digest = getattr(hashlib,type)()
    except Exception:
        raise Exception('cannot get checksum for type %r',type)

    if file and os.path.exists(filename):
        # checksum file contents
        with open(filename,'rb') as filed:
            buffer = filed.read(buffersize)
            while buffer:
                digest.update(buffer)
                buffer = filed.read(buffersize)
    else:
        # just checksum the contents of the first argument
        digest.update(filename)
    return digest.hexdigest()

def md5sum(filename,buffersize=16384):
    """Return md5 digest of file"""
    return cksm(filename,'md5',buffersize)

def sha1sum(filename,buffersize=16384):
    """Return sha1 digest of file"""
    return cksm(filename,'sha1',buffersize)

def sha256sum(filename,buffersize=16384):
    """Return sha256 digest of file"""
    return cksm(filename,'sha256',buffersize)

def sha512sum(filename,buffersize=16384):
    """Return sha512 digest of file"""
    return cksm(filename,'sha512',buffersize)

def load_cksm(sumfile, base_filename):
    """Load the checksum from a file"""
    for l in open(sumfile,'r'):
        if os.path.basename(base_filename) in l:
            sum_cksm, name = l.strip('\n').split()
            return sum_cksm
    raise Exception('could not find checksum in file')

def check_cksm(file,type,sum):
    """Check a checksum of a file"""
    if not os.path.exists(file):
        return False
    # get checksum from file
    file_cksm = cksm(file,type)
    # load sum
    if os.path.isfile(sum):
        sum_cksm = load_cksm(sum, file)
    else:
        sum_cksm = sum
    # check sum
    logging.debug('file_cksm: %r', file_cksm)
    logging.debug('sum_cksm:  %r', sum_cksm)
    return (file_cksm == sum_cksm)

def check_md5sum(file,sum):
    """Check an md5sum of a file"""
    return check_cksm(file,'md5',sum)

def check_sha1sum(file,sum):
    """Check an sha1sum of a file"""
    return check_cksm(file,'sha1',sum)

def check_sha256sum(file,sum):
    """Check an sha256sum of a file"""
    return check_cksm(file,'sha256',sum)

def check_sha512sum(file,sum):
    """Check an sha512sum of a file"""
    return check_cksm(file,'sha512',sum)


### File and Directory Manipulation Functions ###

def removedirs(path):
    try:
        if os.path.isdir(path):
            shutil.rmtree(path,True)
        else:
            os.remove(path)
    except Exception:
        pass

def copy(src,dest):
    parent_dir = os.path.dirname(dest)
    if not os.path.exists(parent_dir):
        logging.info('attempting to make parent dest dir %s',parent_dir)
        try:
            os.makedirs(parent_dir)
        except Exception:
            logging.error('failed to make dest directory for copy',exc_info=True)
            raise
    if os.path.isdir(src):
        logging.info('dircopy: %s to %s',src,dest)
        shutil.copytree(src,dest,symlinks=True)
    else:
        logging.info('filecopy: %s to %s',src,dest)
        shutil.copy2(src,dest)


### Network Functions ###

def getInterfaces():
    """
    Get a list of available interfaces.

    Requires `psutil`.

    Returns:
        dict of {nic_name: {type: address}}
    """
    interfaces = {}

    ret = psutil.net_if_addrs()
    for nic_name in ret:
        n = {}
        for snic in ret[nic_name]:
            if not snic.address:
                continue
            if snic.family == socket.AF_INET:
                n['ipv4'] = snic.address
            elif snic.family == socket.AF_INET6:
                n['ipv6'] = snic.address
            elif snic.family == psutil.AF_LINK:
                n['mac'] = snic.address
        interfaces[nic_name] = n
    return interfaces

def get_local_ip_address():
    """Get the local (loopback) ip address"""
    try:
        return socket.gethostbyname('localhost')
    except Exception:
        return socket.gethostbyname( socket.getfqdn() )

def gethostname():
    """Get hostname of this computer."""
    ret = socket.getfqdn()
    try:
        resp = requests.get('http://simprod.icecube.wisc.edu/downloads/getip.php', timeout=1)
        resp.raise_for_status()
        logging.info('getip: %r', resp.text)
        ret2 = resp.text.split(' ')[-1]
        if len(ret2.split('.')) > 1:
            ret = '.'.join(ret.split('.')[:1]+ret2.split('.')[1:])
    except Exception:
        logging.info('error getting global ip', exc_info=True)
    return ret

@contextmanager
def _http_helper(options={}, sync=True):
    """Set up an http session using requests"""
    if sync:
        session = Session
    else:
        session = AsyncSession
    with session() as s:
        if 'username' in options and 'password' in options:
            s.auth = (options['username'], options['password'])
        if 'sslcert' in options:
            if 'sslkey' in options:
                s.cert = (options['sslcert'], options['sslkey'])
            else:
                s.cert = options['sslcert']
        if 'cacert' in options:
            s.verify = options['cacert']
        yield s

async def download(url, local, options={}):
    """Download a file, checksumming if possible"""
    local = os.path.expanduser(os.path.expandvars(local))
    url  = os.path.expanduser(os.path.expandvars(url))
    if not isurl(url):
        if os.path.exists(url):
            url = 'file:'+url
        else:
            raise Exception("unsupported protocol %s" % url)

    # strip off query params
    if '?' in url:
        clean_url = url[:url.find('?')]
    elif '#' in url:
        clean_url = url[:url.find('#')]
    else:
        clean_url = url

    # fix local to be a filename to write to
    if local.startswith('file:'):
        local = local[5:]
    if os.path.isdir(local):
        local = os.path.join(local, os.path.basename(clean_url))

    logging.warning('wget(): src: %s, local: %s', url, local)

    # actually download the file
    try:
        if url.startswith('http'):
            logging.info('http from %s to %s', url, local)
            # http_proxy fix
            for k in os.environ:
                if k.lower() == 'http_proxy' and not os.environ[k].startswith('http'):
                    os.environ[k] = 'http://'+os.environ[k]
            def _d():
                with _http_helper(options) as s:
                    r = s.get(url, stream=True, timeout=300)
                    with open(local, 'wb') as f:
                        for chunk in r.iter_content(65536):
                            f.write(chunk)
                    r.raise_for_status()
            await asyncio.get_event_loop().run_in_executor(None, _d)
        elif url.startswith('file:'):
            url = url[5:]
            logging.info('copy from %s to %s', url, local)
            if os.path.exists(url):
                await asyncio.get_event_loop().run_in_executor(None,
                        partial(copy, url, local))
        elif url.startswith('gsiftp:') or url.startswith('ftp:'):
            logging.info('gsiftp from %s to %s', url, local)
            await asyncio.get_event_loop().run_in_executor(None,
                    partial(GridFTP.get, url, filename=local))
        else:
            raise Exception("unsupported protocol %s" % url)

        if not os.path.exists(local):
            raise Exception('download failed - file does not exist')
    except Exception:
        await asyncio.get_event_loop().run_in_executor(None, removedirs, local)
        raise

    return local

async def upload(local, url, checksum=True, options={}):
    """Upload a file, checksumming if possible"""
    local = os.path.expandvars(local)
    url  = os.path.expandvars(url)
    if not isurl(url):
        if url.startswith('/'):
            url = 'file:'+url
        else:
            raise Exception("unsupported protocol %s" % url)
    if local.startswith('file:'):
        local = local[5:]
    if os.path.isdir(local):
        compress(local, 'tar')
        local += '.tar'

    logging.warning('wput(): local: %s, url: %s', local, url)

    if not os.path.exists(local):
        logging.warning('upload: local path, %s, does not exist', local)
        raise Exception('local file does not exist')

    chksum = sha512sum(local)
    chksum_type = 'sha512'
    if not checksum:
        logging.warning(f'not performing checksum {chksum_type}\n{checksum}: {url}')

    # actually upload the file
    if url.startswith('http'):
        logging.info('http from %s to %s', local, url)
        # http_proxy fix
        for k in os.environ:
            if k.lower() == 'http_proxy' and not os.environ[k].startswith('http'):
                os.environ[k] = 'http://'+os.environ[k]
        def _d():
            with _http_helper(options) as s:
                with open(local, 'rb') as f:
                    m = MultipartEncoder(
                        fields={'field0': ('filename', f, 'text/plain')}
                        )
                    r = s.post(url, timeout=300, data=m,
                               headers={'Content-Type': m.content_type})
                    r.raise_for_status()
                    if checksum: # get checksum
                        r = s.get(url, stream=True, timeout=300)
                        try:
                            with open(local+'.tmp', 'wb') as f:
                                for chunk in r.iter_content(65536):
                                    f.write(chunk)
                            r.raise_for_status()
                            if sha512sum(local+'.tmp') != chksum:
                                raise Exception('http checksum error')
                        finally:
                            removedirs(local+'.tmp')
        await asyncio.get_event_loop().run_in_executor(None, _d)
    elif url.startswith('file:'):
        # use copy command
        url = url[5:]
        def _c():
            if os.path.exists(url):
                logging.warning('put: file already exists. overwriting!')
                removedirs(url)
            copy(local, url)
            if checksum and sha512sum(url) != chksum:
                raise Exception('file checksum error')
        await asyncio.get_event_loop().run_in_executor(None, _c)
    elif url.startswith('gsiftp:') or url.startswith('ftp:'):
        def _g():
            try:
                GridFTP.put(url, filename=local)
            except Exception:
                # because d-cache doesn't allow overwriting, try deletion
                GridFTP.delete(url)
                GridFTP.put(url, filename=local)
            if checksum and GridFTP.sha512sum(url) != chksum:
                raise Exception('gridftp checksum error')
        await asyncio.get_event_loop().run_in_executor(None, _g)
    else:
        raise Exception("unsupported protocol %s" % url)

def delete(url, options={}):
    """Delete a url or file"""
    url = os.path.expandvars(url)
    if (not isurl(url)) and os.path.exists(url):
        url = 'file:'+url

    if url.startswith('http'):
        logging.info('delete http: %s', url)
        with _http_helper(options) as s:
            r = s.delete(url, timeout=300)
            r.raise_for_status()
    elif url.startswith('file:'):
        url = url[5:]
        logging.info('delete file: %r', url)
        if os.path.exists(url):
            removedirs(url)
    elif url.startswith('gsiftp:') or url.startswith('ftp:'):
        logging.info('delete gsiftp: %r', url)
        GridFTP.rmtree(url)
    else:
        raise Exception("unsupported protocol %s" % url)

def isurl(url):
    """Determine if this is a supported protocol"""
    prefixes = ('file:','http:','https:','ftp:','ftps:','gsiftp:')
    try:
        return url.startswith(prefixes)
    except Exception:
        try:
            return reduce(lambda a,b: a or url.startswith(b), prefixes, False)
        except Exception:
            return False



"""
Server
======

Run the iceprod server.
"""

from __future__ import absolute_import, division, print_function

import os
import sys
import logging
from functools import partial
import importlib
import subprocess
from datetime import datetime, timedelta
import asyncio

from tornado.ioloop import IOLoop

from concurrent.futures import ThreadPoolExecutor

from iceprod.core.logger import set_log_level
from iceprod.server.config import IceProdConfig


logger = logging.getLogger('Server')

class Server(object):
    """
    The actual server.

    """
    def __init__(self, config_params=None, outfile=None, errfile=None):
        self.io_loop = IOLoop.current()
        self.executor = ThreadPoolExecutor(max_workers=10)
        self.cfg = IceProdConfig(override=config_params)
        self.modules = {}
        self.services = {'daemon': {'restart': self.restart,
                                    'reload': self.reload,
                                    'stop': self.stop,
                                    'kill': self.kill,
                                    'get_running_modules': lambda: self.modules.keys(),
                                   },
                        }
        self.outfile = outfile
        self.errfile = errfile

        set_log_level(self.cfg['logging']['level'])

        if 'blocking_threshold' in self.cfg['logging']:
            self.io_loop.set_blocking_log_threshold(self.cfg['logging']['blocking_threshold'])

        for mod_name in self.cfg['modules']:
            if self.cfg['modules'][mod_name]:
                try:
                    m = importlib.import_module('iceprod.server.modules.'+mod_name)
                    mod = getattr(m, mod_name)(cfg=self.cfg,
                                               io_loop=self.io_loop,
                                               executor=self.executor,
                                               modules=self.services)
                    self.modules[mod_name] = mod
                    self.services[mod_name] = mod.service
                    mod.start()
                except Exception:
                    logger.critical('cannot start module', exc_info=True)
                    self.kill()
                    raise

    async def rotate_logs(self):
        current_date = datetime.utcnow()
        while self.outfile and self.errfile:
            if current_date.day != datetime.utcnow().day:
                # rotate files
                current_date = datetime.utcnow()
                if self.outfile:
                    roll_files(sys.stdout, self.outfile)
                if self.errfile:
                    roll_files(sys.stderr, self.errfile)
            await asyncio.sleep(3600)

    def run(self):
        try:
            self.io_loop.add_callback(self.rotate_logs)
            self.io_loop.start()
        except Exception:
            logger.critical('exception not caught', exc_info=True)
            self.kill()

    def restart(self):
        env = os.environ.copy()
        extra_path = os.path.join(os.environ['I3PROD'],'bin')
        env['PATH'] = extra_path+(':'+env['PATH'] if 'PATH' in env else '')
        subprocess.Popen(['iceprod_server.py','restart'],
                         cwd=os.environ['I3PROD'], env=env)

    def reload(self):
        for m in self.modules.values():
            m.stop()
            m.start()

    def stop(self):
        for m in self.modules.values():
            m.stop()
        self.io_loop.stop()

    def kill(self):
        for m in self.modules.values():
            m.kill()
        self.io_loop.stop()

def roll_files(fd, filename, num_files=5):
    d = datetime.utcnow()
    ext = (d-timedelta(days=num_files-1)).strftime('%Y-%m-%d')
    newfile = f'{filename}.{ext}'
    if os.path.exists(newfile): # delete last file
        os.remove(newfile)
    for i in range(num_files-2, 0, -1):
        ext = (d-timedelta(days=i)).strftime('%Y-%m-%d')
        oldfile = f'{filename}.{ext}'
        if os.path.exists(oldfile):
            os.rename(oldfile, newfile)
        newfile = oldfile
    if os.path.exists(filename):
        os.rename(filename, newfile)

    # redirect file descriptor
    newfd = open(filename, fd.mode)
    fd.flush()
    os.dup2(newfd.fileno(), fd.fileno())
    if fd != sys.stdout and fd != sys.stderr:
        fd.close()
    return newfd



import logging
import json
import uuid

import tornado.web
import pymongo
import motor

from iceprod.server.rest import RESTHandler, RESTHandlerSetup, authorization
from iceprod.server.util import nowstr

logger = logging.getLogger('rest.task_stats')

def setup(config, *args, **kwargs):
    """
    Setup method for TaskStats REST API.

    Sets up any database connections or other prerequisites.

    Args:
        config (dict): an instance of :py:class:`iceprod.server.config`.

    Returns:
        list: Routes for logs, which can be passed to :py:class:`tornado.web.Application`.
    """
    cfg_rest = config.get('rest',{}).get('task_stats',{})
    db_cfg = cfg_rest.get('database',{})

    # add indexes
    db = pymongo.MongoClient(**db_cfg).task_stats
    if 'task_stat_id_index' not in db.task_stats.index_information():
        db.task_stats.create_index('task_stat_id', name='task_stat_id_index', unique=True)
    if 'task_id_index' not in db.task_stats.index_information():
        db.task_stats.create_index('task_id', name='task_id_index', unique=False)
    if 'dataset_id_index' not in db.task_stats.index_information():
        db.task_stats.create_index('dataset_id', name='dataset_id_index', unique=False)

    handler_cfg = RESTHandlerSetup(config, *args, **kwargs)
    handler_cfg.update({
        'database': motor.motor_tornado.MotorClient(**db_cfg).task_stats,
    })

    return [
        (r'/tasks/(?P<task_id>\w+)/task_stats', MultiTaskStatsHandler, handler_cfg),
        (r'/datasets/(?P<dataset_id>\w+)/tasks/(?P<task_id>\w+)/task_stats', DatasetsMultiTaskStatsHandler, handler_cfg),
        (r'/datasets/(?P<dataset_id>\w+)/bulk/task_stats', DatasetsBulkTaskStatsHandler, handler_cfg),
        (r'/datasets/(?P<dataset_id>\w+)/tasks/(?P<task_id>\w+)/task_stats/(?P<task_stat_id>\w+)', DatasetsTaskStatsHandler, handler_cfg),
    ]

class BaseHandler(RESTHandler):
    def initialize(self, database=None, **kwargs):
        super(BaseHandler, self).initialize(**kwargs)
        self.db = database

class MultiTaskStatsHandler(BaseHandler):
    """
    Handle multi task_stats requests.
    """
    @authorization(roles=['admin','client','pilot'])
    async def post(self, task_id):
        """
        Create a task_stat entry.

        Body should contain the task stat data.

        Args:
            task_id (str): the task id for this task_stat

        Returns:
            dict: {'result': <task_stat_id>}
        """
        stat_data = json.loads(self.request.body)
        if 'dataset_id' not in stat_data:
            raise tornado.web.HTTPError(400, reason='Missing dataset_id in body')

        # set some fields
        task_stat_id = uuid.uuid1().hex
        data = {
            'task_stat_id': task_stat_id,
            'task_id': task_id,
            'dataset_id': stat_data['dataset_id'],
            'create_date': nowstr(),
            'stats': stat_data,
        }

        ret = await self.db.task_stats.insert_one(data)
        self.set_status(201)
        self.write({'result': task_stat_id})
        self.finish()

class DatasetsBulkTaskStatsHandler(BaseHandler):
    """
    Handle a dataset bulk task_stats requests.

    Stream the output of all stats.
    """
    @authorization(roles=['admin'], attrs=['dataset_id:read'])
    async def get(self, dataset_id):
        """
        Get task_stats for a dataset and task.

        Args:
            dataset_id (str): dataset id

        Params (optional):
            last: bool (True: only last task_stat.  False: all task_stats)
            after: only return stats created more recently than this date
            keys: | separated list of keys to return for each task_stat
            buffer_size: number of records to buffer before flushing (default 100)

        Returns:
            dict: {<task_stat_id>: stats}
        """
        last = self.get_argument('last', 'f').lower() in ('true','t','1','yes','y')

        query = {'dataset_id':dataset_id}
        after = self.get_argument('after', None)
        if after:
            query['create_date'] = {"$gte": after}

        projection = {'_id': False}
        keys = self.get_argument('keys','')
        if keys:
            keys = keys.split('|')
            projection.update({x:True for x in keys if x})
            projection['task_stat_id'] = True
            projection['task_id'] = True
            projection['create_date'] = True
        buffer_size = int(self.get_argument('buffer_size', '1000'))

        task_id = None
        data = []
        n = 0
        async for row in self.db.task_stats.find(query, projection=projection).sort([('task_id',1)]):
            if row['task_id'] == task_id:
                data.append(row)
                continue
            if data:
                ret = sorted(data, key=lambda x: x['create_date'])
                if keys:
                    ret = [{k:d[k] for k in d if k in keys} for d in ret]
                if last:
                    self.write(ret[-1])
                    self.write('\n')
                else:
                    for ret in data:
                        self.write(ret)
                        self.write('\n')
                n += 1
                if n >= buffer_size:
                    n = 0
                    await self.flush()
            data = [row]
            task_id = row['task_id']

        if data:
            ret = sorted(data, key=lambda x: x['create_date'])
            if keys:
                ret = [{k:d[k] for k in d if k in keys} for d in ret]
            if last:
                self.write(ret[-1])
                self.write('\n')
            else:
                for ret in data:
                    self.write(ret)
                    self.write('\n')
        self.finish()

class DatasetsMultiTaskStatsHandler(BaseHandler):
    """
    Handle multi task_stats requests.
    """
    @authorization(roles=['admin'], attrs=['dataset_id:read'])
    async def get(self, dataset_id, task_id):
        """
        Get task_stats for a dataset and task.

        Args:
            dataset_id (str): dataset id
            task_id (str): task id

        Params (optional):
            last: bool (True: only last task_stat.  False: all task_stats)
            keys: | separated list of keys to return for each task_stat

        Returns:
            dict: {<task_stat_id>: stats}
        """
        last = self.get_argument('last', 'f').lower() in ('true','t','1','yes','y')

        projection = {'_id': False}
        keys = self.get_argument('keys','')
        if keys:
            projection.update({x:True for x in keys.split('|') if x})
            projection['task_stat_id'] = True
            if last:
                projection['create_date'] = True

        ret = await self.db.task_stats.find({'dataset_id':dataset_id,'task_id':task_id},
                projection=projection).to_list(10000)

        if last:
            ret = sorted(ret, key=lambda x: x['create_date'])[-1:]

        self.write({row['task_stat_id']:row for row in ret})
        self.finish()

class DatasetsTaskStatsHandler(BaseHandler):
    """
    Handle single task_stat requests.
    """
    @authorization(roles=['admin'], attrs=['dataset_id:read'])
    async def get(self, dataset_id, task_id, task_stat_id):
        """
        Get a task_stat entry.

        Args:
            dataset_id (str): dataset id
            task_id (str): task id
            task_stat_id (str): the task_stat id

        Returns:
            dict: task_stat entry
        """
        ret = await self.db.task_stats.find_one({'dataset_id':dataset_id,'task_id':task_id,'task_stat_id':task_stat_id},
                projection={'_id':False})
        if not ret:
            self.send_error(404, reason="Task stat not found")
        else:
            self.write(ret)
            self.finish()



"""
Materialization service for dataset late materialization.
"""
import logging
import importlib
import os
import time
from datetime import datetime, timedelta
import json
import uuid
import asyncio
from collections import defaultdict

import pymongo
import motor
import motor.motor_asyncio
from rest_tools.server import RestServer
from rest_tools.client import RestClient

import iceprod.server
from iceprod.server import module
from iceprod.server import get_pkgdata_filename
from iceprod.server.rest import RESTHandler, RESTHandlerSetup, authorization
from iceprod.server.util import nowstr, datetime2str, task_statuses
from iceprod.core.parser import ExpParser
from iceprod.core.resources import Resources
from iceprod.server.priority import Priority

logger = logging.getLogger('modules_materialization')

class materialization(module.module):
    """
    Run the late materialization module, which handles a REST service for
    dataset late materialization.

    """
    def __init__(self,*args,**kwargs):
        super(materialization,self).__init__(*args,**kwargs)

        # set up materialization service
        ms = MaterializationService(self.cfg)

        # set up the REST API
        routes, args = setup_rest(self.cfg, module=self, materialization_service=ms)
        self.server = RestServer(**args)
        for r in routes:
            self.server.add_route(*r)

        kwargs = {}
        if 'materialization' in self.cfg:
            if 'address' in self.cfg['materialization']:
                kwargs['address'] = self.cfg['materialization']['address']
            if 'port' in self.cfg['materialization']:
                kwargs['port'] = self.cfg['materialization']['port']

        self.server.startup(**kwargs)


class MaterializationService:
    """Materialization service."""
    def __init__(self, cfg):
        self.cfg = cfg

        db_cfg = cfg.get('materialization',{}).get('database',{})
        self.db = motor.motor_asyncio.AsyncIOMotorClient(**db_cfg).datasets

        rest_cfg = cfg.get('rest_api', {})
        rest_url = rest_cfg.get('url', 'https://iceprod2-api.icecube.wisc.edu')
        if 'auth_key' not in rest_cfg:
            raise Exception('no auth key for rest api')
        rest_auth_key = rest_cfg.get('auth_key')
        rest_client = RestClient(rest_url, rest_auth_key)
        self.materialize = Materialize(rest_client)

        self.start_time = time.time()
        self.last_run_time = None
        self.last_cleanup_time = None
        self.last_success_time = None
        asyncio.get_event_loop().create_task(self.run())

    async def run(self):
        """
        Run loop.
        """
        while True:
            ret = None
            try:
                self.last_run_time = time.time()
                now = nowstr()

                # periodically cleanup
                if (not self.last_cleanup_time) or self.last_run_time - self.last_cleanup_time > 3600*6:
                    clean_time = datetime2str(datetime.utcfromtimestamp(self.last_run_time)-timedelta(hours=6))
                    await self.db.materialization.delete_many({'status': {'$in': ['complete', 'error']}, 'modify_timestamp': {'$lt': clean_time}})
                    await self.db.materialization.update_many({'status': 'processing', 'modify_timestamp': {'$lt': clean_time}}, {'$set': {'status': 'waiting', 'modify_timestamp': now}})
                    self.last_cleanup_time = time.time()

                # get next materialization from DB
                ret = await self.db.materialization.find_one_and_update(
                        {'status': 'waiting'},
                        {'$set': {'status': 'processing', 'modify_timestamp': now}},
                        projection={'_id':False},
                        sort=[('modify_timestamp', 1)],
                        return_document=pymongo.ReturnDocument.AFTER,
                )
                if not ret:
                    logger.info('materialization service has nothing to do. sleeping')
                    await asyncio.sleep(60)
                    continue

                # run materialization
                logger.warning(f'running materialization request {ret["materialization_id"]}')
                kwargs = {}
                if 'dataset_id' in ret and ret['dataset_id']:
                    kwargs['only_dataset'] = ret['dataset_id']
                if 'num' in ret and ret['num']:
                    kwargs['num'] = ret['num']
                if 'set_status' in ret and ret['set_status']:
                    kwargs['set_status'] = ret['set_status']
                await self.materialize.run_once(**kwargs)

                await self.db.materialization.update_one(
                        {'materialization_id': ret['materialization_id']},
                        {'$set': {'status': 'complete'}},
                )
                self.last_success_time = time.time()
            except Exception:
                logger.error('error running materialization', exc_info=True)
                if ret:
                    await self.db.materialization.update_one(
                            {'materialization_id': ret['materialization_id']},
                            {'$set': {'status': 'error'}},
                    )


class Materialize:
    def __init__(self, rest_client):
        self.rest_client = rest_client
        self.config_cache = {}
        self.prio = None

    async def run_once(self, only_dataset=None, set_status=None, num=20000, dryrun=False):
        """
        Actual materialization work.

        Args:
            only_dataset (str): dataset_id if we should only buffer a single dataset
            set_status (str): status of new tasks
            num (int): max number of jobs to buffer
            dryrun (bool): if true, do not modify DB, just log changes
        """
        if set_status and set_status not in task_statuses:
            raise Exception('set_status is not a valid task status')
        self.config_cache = {} # clear config cache
        self.prio = Priority(self.rest_client) # clear priority cache
        datasets = await self.rest_client.request('GET', '/dataset_summaries/status')
        if 'truncated' in datasets and only_dataset:
            datasets['processing'].extend(datasets['truncated'])
        if 'suspended' in datasets and only_dataset:
            datasets['processing'].extend(datasets['suspended'])
        if 'processing' in datasets:
            for dataset_id in datasets['processing']:
                if only_dataset and dataset_id != only_dataset:
                    continue
                try:
                    dataset = await self.rest_client.request('GET', '/datasets/{}'.format(dataset_id))
                    job_counts = await self.rest_client.request('GET', '/datasets/{}/job_counts/status'.format(dataset_id))
                    tasks = await self.rest_client.request('GET', '/datasets/{}/task_counts/status'.format(dataset_id))
                    if 'waiting' not in tasks or 'processing' not in job_counts or job_counts['processing'] < num or only_dataset:
                        # buffer for this dataset
                        logger.warning('checking dataset %s', dataset_id)
                        jobs = await self.rest_client.request('GET', '/datasets/{}/jobs'.format(dataset_id), {'keys': 'job_id|job_index'})

                        # check that last job was buffered correctly
                        job_index = max(jobs[i]['job_index'] for i in jobs)+1 if jobs else 0
                        num_tasks = sum(tasks.values())
                        while num_tasks % dataset['tasks_per_job'] != 0 and job_index > 0:
                            # a job must have failed to buffer, so check in reverse order
                            job_index -= 1
                            job_tasks = await self.rest_client.request('GET', f'/datasets/{dataset_id}/tasks',
                                                                       {'job_index': job_index, 'keys': 'task_id|job_id|task_index'})
                            if len(job_tasks) != dataset['tasks_per_job']:
                                logger.info('fixing buffer of job %d for dataset %s', job_index, dataset_id)
                                ret = await self.rest_client.request('GET', f'/datasets/{dataset_id}/jobs',
                                                                     {'job_index': job_index, 'keys': 'job_id'})
                                job_id = list(ret.keys())[0]
                                tasks_buffered = await self.buffer_job(dataset, job_index, job_id=job_id,
                                                                       tasks=list(job_tasks.values()),
                                                                       set_status=set_status, dryrun=dryrun)
                                num_tasks += tasks_buffered

                        # now try buffering new tasks
                        job_index = max(jobs[i]['job_index'] for i in jobs)+1 if jobs else 0
                        jobs_to_buffer = min(num, dataset['jobs_submitted'] - len(jobs))
                        if jobs_to_buffer > 0:
                            logger.info('buffering %d jobs for dataset %s', jobs_to_buffer, dataset_id)
                            for i in range(jobs_to_buffer):
                                await self.buffer_job(dataset, job_index, set_status=set_status, dryrun=dryrun)
                                job_index += 1
                except Exception:
                    logger.error('error buffering dataset %s', dataset_id, exc_info=True)
                    if only_dataset:
                        raise

    async def buffer_job(self, dataset, job_index, job_id=None, tasks=None, set_status=None, dryrun=False):
        """
        Buffer a single job for a dataset

        Args:
            dataset (dict): dataset info
            job_index (int): job index
            job_id (str): job id (if filling in remaining)
            tasks (list): existing tasks (if filling in remaining)
            set_status (str): status of new tasks
            dryrun (bool): set to True if this is a dry run

        Returns:
            int: number of tasks buffered
        """
        dataset_id = dataset['dataset_id']
        logger.info('buffering dataset %s job %d', dataset_id, job_index)

        config = await self.get_config(dataset_id)
        parser = ExpParser()
        task_names = [task['name'] if task['name'] else str(i) for i,task in enumerate(config['tasks'])]
        if len(task_names) != dataset['tasks_per_job']:
            raise Exception('config num tasks does not match dataset tasks_per_job')

        args = {'dataset_id': dataset_id, 'job_index': job_index}
        if dryrun:
            job_id = {'result': 'DRYRUN'}
            task_ids = []
            task_iter = enumerate(task_names)
        elif job_id:
            task_ids = [task['task_id'] for task in tasks] if tasks else []
            task_indexes = {task['task_index'] for task in tasks} if tasks else {}
            task_iter = [(i,name) for i,name in enumerate(task_names) if i not in task_indexes]
            if not task_iter:
                raise Exception('no task names to create')
        else:
            ret = await self.rest_client.request('POST', '/jobs', args)
            job_id = ret['result']
            task_ids = []
            task_iter = enumerate(task_names)

        # buffer tasks
        for task_index,name in task_iter:
            depends = await self.get_depends(config, job_index,
                                             task_index, task_ids)
            config['options']['job'] = job_index
            config['options']['task'] = task_index
            config['options']['dataset'] = dataset['dataset']
            config['options']['jobs_submitted'] = dataset['jobs_submitted']
            config['options']['tasks_submitted'] = dataset['tasks_submitted']
            config['options']['debug'] = dataset['debug']
            args = {
                'dataset_id': dataset_id,
                'job_id': job_id,
                'task_index': task_index,
                'job_index': job_index,
                'name': name,
                'depends': depends,
                'requirements': self.get_reqs(config, task_index, parser),
            }
            if set_status:
                args['status'] = set_status
            if dryrun:
                logger.info(f'DRYRUN: POST /tasks {args}')
                task_ids.append(job_index*1000000+task_index)
            else:
                ret = await self.rest_client.request('POST', '/tasks', args)
                task_id = ret['result']
                task_ids.append(task_id)
                p = await self.prio.get_task_prio(dataset_id, task_id)
                await self.rest_client.request('PATCH', f'/tasks/{task_id}', {'priority': p})

        return len(task_ids)

    async def get_config(self, dataset_id):
        """Get dataset config"""
        if dataset_id in self.config_cache:
            return self.config_cache[dataset_id]

        config = await self.rest_client.request('GET', '/config/{}'.format(dataset_id))
        if 'options' not in config:
            config['options'] = {}
        self.config_cache[dataset_id] = config
        return config

    def get_reqs(self, config, task_index, parser):
        """
        Get requirements for a task.

        Args:
            config (:py:class:`iceprod.core.dataclasses.Job`): dataset config
            task_index (int): task index
            parser (:py:class:`iceprod.core.parser.ExpParser`): parser

        Returns:
            dict: task requirements
        """
        task = config['tasks'][task_index]
        req = task['requirements'].copy() if 'requirements' in task else {}
        for k in req:
            if isinstance(req[k], list):
                req[k] = [parser.parse(val,config) for val in req[k]]
            else:
                req[k] = parser.parse(req[k],config)
        for k in Resources.defaults:
            # don't add if not needed?
            #if k not in req or not req[k]:
            #    req[k] = Resources.defaults[k]
            if k == 'gpu' and k in req and isinstance(req[k], (tuple,list)):
                req[k] = len(req[k])
        return req

    async def get_depends(self, config, job_index, task_index, task_ids):
        """
        Get dependency task_ids for a task.

        Args:
            config (:py:class:`iceprod.core.dataclasses.Job`): dataset config
            job_index (int): job index
            task_index (int): task index
            task_ids (list): list of already buffered task_ids in this job

        Returns:
            list: list of task_id dependencies
        """
        task = config['tasks'][task_index]
        if 'depends' not in task:
            return []
        all_tasks ={task['name'] if task['name'] else str(i):i for i,task in enumerate(config['tasks'])}
        ret = []
        for dep in task['depends']:
            if dep in all_tasks:
                logging.debug('dep %r in all_tasks', dep)
                dep_index = all_tasks[dep]
                if dep_index >= task_index:
                    raise Exception("bad depends: %r. can't depend on ourself or a greater index", dep)
                ret.append(task_ids[dep_index])
            elif isinstance(dep, int) and dep < len(all_tasks):
                logging.debug('dep %r is int', dep)
                dep_index = dep
                if dep_index >= task_index:
                    raise Exception("bad depends: %r. can't depend on ourself or a greater index", dep)
                ret.append(task_ids[dep_index])
            elif isinstance(dep,str) and ':' in dep:
                # try for a task in another dataset
                logging.debug('dep %r in another dataset', dep)
                dataset_id, dep = dep.split(':',1)
                try:
                    #tasks = await rest_client.request('GET', '/datasets/{}/tasks?keys=task_id|name|task_index&job_index={}'.format(dataset_id,job_index))
                    tasks = await self.rest_client.request('GET', '/datasets/{}/tasks?keys=task_id|name|task_index'.format(dataset_id))
                    for task in tasks.values():
                        if dep == task['name'] or dep == str(task['task_index']):
                            job = await self.rest_client.request('GET', '/jobs/{}'.format(task['job_id']))
                            logging.info('ext job_index=%r, my job_index=%r', job['job_index'], job_index)
                            if job['job_index'] == job_index:
                                ret.append(task['task_id'])
                                break
                    else:
                        raise Exception()
                except Exception:
                    raise Exception('bad depends: %r', dep)
            else:
                # try for a raw task_id
                logging.debug('dep %r is a raw id', dep)
                try:
                    task = await self.rest_client.request('GET', '/tasks/{}'.format(dep))
                    ret.append(task['task_id'])
                except Exception:
                    raise Exception('bad depends: %r', dep)

        return ret


def setup_rest(config, materialization_service=None, **kwargs):
    """
    Setup a REST Tornado server.

    Args:
        config (:py:class:`iceprod.server.config.Config`): An IceProd config

    Returns:
        tuple: (routes, application args)
    """
    cfg_rest = config.get('materialization',{})
    db_cfg = cfg_rest.get('database',{})

    # add indexes
    db = pymongo.MongoClient(**db_cfg).datasets
    if 'materialization_id_index' not in db.materialization.index_information():
        db.materialization.create_index('materialization_id', name='materialization_id_index', unique=True)
    if 'dataset_id_index' not in db.materialization.index_information():
        db.materialization.create_index('dataset_id', name='dataset_id_index')
    if 'status_timestamp_index' not in db.materialization.index_information():
        db.materialization.create_index([('status',pymongo.ASCENDING),('timestamp',pymongo.ASCENDING)], name='status_timestamp_index')

    handler_cfg = RESTHandlerSetup(config, **kwargs)
    handler_cfg['database'] = motor.motor_tornado.MotorClient(**db_cfg).datasets
    handler_cfg['materialization_service'] = materialization_service

    routes = [
        (r'/status/(?P<materialization_id>\w+)', StatusHandler, handler_cfg),
        (r'/', RequestHandler, handler_cfg),
        (r'/request/(?P<dataset_id>\w+)', RequestDatasetHandler, handler_cfg),
        (r'/healthz', HealthHandler, handler_cfg),
    ]

    logger.info('REST routes being served:')
    for r in routes:
        logger.info('  %r', r)

    kwargs = {}
    return (routes, kwargs)

class BaseHandler(RESTHandler):
    """
    Base handler for Materialization REST API.
    """
    def initialize(self, database=None, materialization_service=None, **kwargs):
        super(BaseHandler, self).initialize(**kwargs)
        self.db = database
        self.materialization_service = materialization_service

    async def new_request(self, args):
        # validate first
        fields = {
            'dataset_id': str,
            'set_status': str,
            'num': int,
        }
        if set(args)-set(fields): # don't let random args through
            raise tornado.web.HTTPError(400, reason='invalid params')
        for k in fields:
            if k in args and not isinstance(args[k], fields[k]):
                r = 'key "{}" should be of type {}'.format(k, fields[k].__name__)
                raise tornado.web.HTTPError(400, reason=r)

        # set some fields
        now = nowstr()
        data = {
            'materialization_id': uuid.uuid1().hex,
            'status': 'waiting',
            'create_timestamp': now,
            'modify_timestamp': now,
            'creator': self.auth_data['username'],
            'role': self.auth_data['role'],
        }
        for k in fields:
            if k in args:
                data[k] = args[k]

        # insert
        await self.db.materialization.insert_one(data)
        return data

class StatusHandler(BaseHandler):
    """
    Handle materialization status requests.
    """
    @authorization(roles=['admin','client','user'])
    async def get(self, materialization_id):
        """
        Get materialization status.

        If materialization_id is invalid, returns http code 404.

        Args:
            materialization_id (str): materialization request id

        Returns:
            dict: materialization metadata
        """
        ret = await self.db.materialization.find_one({'materialization_id':materialization_id},
                projection={'_id':False})
        if not ret:
            self.send_error(404, reason="Materialization request not found")
        else:
            self.write(ret)

class RequestHandler(BaseHandler):
    """
    Handle basic materialization requests.
    """
    @authorization(roles=['admin','client'])
    async def post(self):
        """
        Create basic materialization request.

        On success, returns http code 201.

        Params:
            num: number of jobs to buffer per dataset

        Returns:
            dict: {result: materialization_id}
        """
        args = json.loads(self.request.body) if self.request.body else {}

        data = await self.new_request(args)

        # return success
        self.set_status(201)
        self.write({'result': data['materialization_id']})

class RequestDatasetHandler(BaseHandler):
    """
    Handle dataset materialization requests.
    """
    @authorization(roles=['admin','client'], attrs=['dataset_id:read'])
    async def post(self, dataset_id):
        """
        Create dataset materialization request.

        On success, returns http code 201.

        Params:
            num: number of jobs to buffer per dataset

        Args:
            dataset_id (str): dataset_id to materialize

        Returns:
            dict: {result: materialization_id}
        """
        args = json.loads(self.request.body) if self.request.body else {}
        args['dataset_id'] = dataset_id

        data = await self.new_request(args)

        # return success
        self.set_status(201)
        self.write({'result': data['materialization_id']})

class HealthHandler(BaseHandler):
    """
    Handle health requests.
    """
    async def get(self):
        """
        Get health status.

        Returns based on exit code, 200 = ok, 400 = failure
        """
        now = time.time()
        status = {
            'now': nowstr(),
            'start_time': datetime2str(datetime.utcfromtimestamp(self.materialization_service.start_time)),
            'last_run_time': "",
            'last_success_time': "",
            'last_cleanup_time': "",
            'num_requests': -1,
        }
        try:
            if self.materialization_service.last_run_time is None and self.materialization_service.start_time + 3600 < now:
                self.send_error(500, reason='materialization was never run')
                return
            if self.materialization_service.last_run_time is not None:
                if self.materialization_service.last_run_time + 3600 < now:
                    self.send_error(500, reason='materialization has stopped running')
                    return
                status['last_run_time'] = datetime2str(datetime.utcfromtimestamp(self.materialization_service.last_run_time))
            if self.materialization_service.last_success_time is None and self.materialization_service.start_time + 86400 < now:
                self.send_error(500, reason='materialization was never successful')
                return
            if self.materialization_service.last_success_time is not None:
                if self.materialization_service.last_success_time + 86400 < now:
                    self.send_error(500, reason='materialization has stopped being successful')
                    return
                status['last_success_time'] = datetime2str(datetime.utcfromtimestamp(self.materialization_service.last_success_time))
            if self.materialization_service.last_cleanup_time is not None:
                status['last_cleanup_time'] = datetime2str(datetime.utcfromtimestamp(self.materialization_service.last_cleanup_time))
        except Exception:
            logger.info('error from materialization service', exc_info=True)
            self.send_error(500, reason='error from materialization service')
            return

        try:
            ret = await self.db.materialization.count_documents({'status':{'$in':['waiting','processing']}}, maxTimeMS=1000)
        except Exception:
            logger.info('bad db request', exc_info=True)
            self.send_error(500, reason='bad db request')
            return
        if ret is None:
            self.send_error(500, reason='bad db result')
        else:
            status['num_requests'] = ret
        self.write(status)


if __name__ == '__main__':
    import argparse
    parser = argparse.ArgumentParser(description='Materialize a dataset')
    parser.add_argument('dataset_id')
    parser.add_argument('--rest_url', default='https://iceprod2-api.icecube.wisc.edu')
    parser.add_argument('-t', '--rest_token', default=None)
    parser.add_argument('--set_status', default=None,help='initial task status')
    parser.add_argument('-n','--num', default=100,type=int,help='number of jobs to materialize')
    parser.add_argument('--job_index', type=int, help='specific job index to buffer')
    parser.add_argument('--job_id', default=None, help='specific job id to buffer tasks into')
    parser.add_argument('--debug',action='store_true')
    parser.add_argument('--dryrun',action='store_true',help='do not modify database, just log changes')
    args = parser.parse_args()
    if not args.rest_token:
        raise Exception('no token for rest api')
    logging.basicConfig(level=(logging.DEBUG if args.debug else logging.INFO))

    rest_client = RestClient(args.rest_url, args.rest_token)
    materialize = Materialize(rest_client)
    if args.job_index:
        logging.warning('manually buffering a job for dataset %s job %d', args.dataset_id, args.job_index)
        async def run():
            dataset = await rest_client.request('GET', '/datasets/{}'.format(args.dataset_id))
            materialize.prio = Priority(rest_client)
            await materialize.buffer_job(dataset, args.job_index, job_id=args.job_id)
        asyncio.run(run())
    else:
        asyncio.run(materialize.run_once(only_dataset=args.dataset_id, set_status=args.set_status, num=args.num, dryrun=args.dryrun))



import logging
import json
import uuid
import time

import tornado.web
import pymongo
import motor
import ldap3

from iceprod.server.rest import RESTHandler, RESTHandlerSetup, authorization, authenticated, catch_error

logger = logging.getLogger('rest.auth')

def setup(config, *args, **kwargs):
    """
    Setup method for Auth REST API.

    Sets up any database connections or other prerequisites.

    Args:
        config (dict): an instance of :py:class:`iceprod.server.config`.

    Returns:
        list: Routes for auth, which can be passed to :py:class:`tornado.web.Application`.
    """
    cfg_rest = config.get('rest',{}).get('auth',{})
    db_cfg = cfg_rest.get('database',{})

    # add indexes
    db = pymongo.MongoClient(**db_cfg).auth
    if 'name_index' not in db.roles.index_information():
        db.roles.create_index('name', name='name_index', unique=True)
    if 'name_index' not in db.groups.index_information():
        db.groups.create_index('name', name='name_index', unique=True)
    if 'user_id_index' not in db.users.index_information():
        db.users.create_index('user_id', name='user_id_index', unique=True)
    if 'username_index' not in db.users.index_information():
        db.users.create_index('username', name='username_index', unique=True)
    if 'dataset_id_index' not in db.auths_dataset.index_information():
        db.auths_dataset.create_index('dataset_id', name='dataset_id_index', unique=True)

    handler_cfg = RESTHandlerSetup(config, *args, **kwargs)
    handler_cfg.update({
        'database': motor.motor_tornado.MotorClient(**db_cfg).auth,
    })
    ldap_cfg = dict(handler_cfg)
    ldap_cfg.update({
        'ldap_uri': cfg_rest.get('ldap_uri', ''),
        'ldap_base': cfg_rest.get('ldap_base', ''),
    })

    return [
        (r'/roles', MultiRoleHandler, handler_cfg),
        (r'/roles/(?P<role_name>\w+)', RoleHandler, handler_cfg),
        (r'/groups', MultiGroupHandler, handler_cfg),
        (r'/groups/(?P<group_name>.*)', GroupHandler, handler_cfg),
        (r'/users', MultiUserHandler, handler_cfg),
        (r'/users/(?P<user_id>\w+)', UserHandler, handler_cfg),
        (r'/users/(?P<user_id>\w+)/roles', UserRolesHandler, handler_cfg),
        (r'/users/(?P<user_id>\w+)/groups', UserGroupsHandler, handler_cfg),
        (r'/ldap', LDAPHandler, ldap_cfg),
        (r'/create_token', CreateTokenHandler, handler_cfg),
        (r'/auths/(?P<dataset_id>\w+)', AuthDatasetHandler, handler_cfg),
        (r'/auths/(?P<dataset_id>\w+)/actions/(?P<action>\w+)', AuthDatasetActionHandler, handler_cfg),
    ]


class AuthHandler(RESTHandler):
    """
    Base handler for Auth REST API. 
    """
    def initialize(self, database=None, **kwargs):
        super(AuthHandler, self).initialize(**kwargs)
        self.db = database

class MultiRoleHandler(AuthHandler):
    """
    Handle multi-role requests.
    """
    @authorization(roles=['admin'])
    async def get(self):
        """
        Get a list of roles.

        Returns:
            dict: {'results': list of roles}
        """
        ret = await self.db.roles.find(projection={'_id':False}).to_list(length=1000)
        self.write({'results':ret})
        self.finish()

class RoleHandler(AuthHandler):
    """
    Handle individual role requests.
    """
    @authorization(roles=['admin'])
    async def put(self, role_name):
        """
        Add/modify a role.

        Body should contain all necessary fields for a role.

        Args:
            role_name (str): the role to add/modify

        Returns:
            dict: empty dict
        """
        data = json.loads(self.request.body)
        if 'name' not in data:
            data['name'] = role_name
        elif data['name'] != role_name:
            raise tornado.web.HTTPError(400, 'role name mismatch')
        ret = await self.db.roles.find_one_and_replace({'name':role_name},
                data, upsert=True)
        self.write({})
        self.finish()

    @authorization(roles=['admin','client'])
    async def get(self, role_name):
        """
        Get a role.

        Args:
            role_name (str): the role to get

        Returns:
            dict: role info
        """
        ret = await self.db.roles.find_one({'name':role_name},
                projection={'_id':False})
        if not ret:
            self.send_error(404, reason="Role not found")
        else:
            self.write(ret)
            self.finish()

    @authorization(roles=['admin'])
    async def delete(self, role_name):
        """
        Delete a role.

        Args:
            role_name (str): the role to delete

        Returns:
            dict: empty dict
        """
        await self.db.roles.delete_one({'name':role_name})
        self.write({})
        self.finish()

class MultiGroupHandler(AuthHandler):
    """
    Handle multi-group requests.
    """
    @authorization(roles=['admin','client'])
    async def get(self):
        """
        Get a list of groups.

        Returns:
            dict: {'results': list of groups}
        """
        ret = await self.db.groups.find(projection={'_id':False}).to_list(length=10000)
        self.write({'results':ret})
        self.finish()

class GroupHandler(AuthHandler):
    """
    Handle individual group requests.
    """
    @authorization(roles=['admin'])
    async def put(self, group_name):
        """
        Add/modify a group.

        Body should contain all necessary fields for a group.

        Args:
            group_name (str): the group to add/modify

        Returns:
            dict: empty dict
        """
        data = json.loads(self.request.body)
        if 'name' not in data:
            data['name'] = group_name
        elif data['name'] != group_name:
            raise tornado.web.HTTPError(400, 'group name mismatch')
        ret = await self.db.groups.find_one_and_replace({'name':group_name},
                data, upsert=True)
        logger.info('%r', ret)
        self.write({})
        self.finish()

    @authorization(roles=['admin','client'])
    async def get(self, group_name):
        """
        Get a group.

        Args:
            group_name (str): the group to get

        Returns:
            dict: group info
        """
        ret = await self.db.groups.find_one({'name':group_name},
                projection={'_id':False})
        if not ret:
            self.send_error(404, reason="Group not found")
        else:
            self.write(ret)
            self.finish()

    @authorization(roles=['admin'])
    async def delete(self, group_name):
        """
        Delete a group.

        Args:
            group_name (str): the group to delete

        Returns:
            dict: empty dict
        """
        await self.db.groups.delete_one({'name':group_name})
        self.write({})
        self.finish()

class MultiUserHandler(AuthHandler):
    """
    Handle multi-user requests.
    """
    @authorization(roles=['admin','client'])
    async def get(self):
        """
        Get a list of users.

        Body args:
            username: username to filter by

        Returns:
            dict: {'results': list of users}
        """
        filters = {}
        try:
            data = json.loads(self.request.body)
            if 'username' in data:
                filters['username'] = data['username']
        except Exception:
            pass

        ret = await self.db.users.find(filters, projection={'_id':False}).to_list(length=1000)
        self.write({'results':ret})
        self.finish()

    @authorization(roles=['admin'])
    async def post(self):
        """
        Add a user.

        Body should contain all necessary fields for a user.
        """
        data = json.loads(self.request.body)
        if 'username' not in data:
            raise tornado.web.HTTPError(400, reason='missing username')
        ret = await self.db.users.find_one({'username': data['username']})
        if ret:
            raise tornado.web.HTTPError(400, reason='duplicate username')

        user_id = uuid.uuid1().hex
        data['user_id'] = user_id
        if 'groups' not in data:
            data['groups'] = []
        if 'roles' not in data:
            data['roles'] = []
        if 'priority' not in data:
            data['priority'] = 0.5
        ret = await self.db.users.insert_one(data)
        self.set_status(201)
        self.set_header('Location', f'/users/{user_id}')
        self.write({'result': f'/users/{user_id}'})
        self.finish()

class UserHandler(AuthHandler):
    """
    Handle individual user requests.
    """
    @authorization(roles=['admin','client'])
    async def get(self, user_id):
        """
        Get a user.

        Args:
            user_id (str): the user to get

        Returns:
            dict: user info
        """
        ret = await self.db.users.find_one({'user_id':user_id},
                projection={'_id':False})
        if not ret:
            self.send_error(404, reason="User not found")
        else:
            self.write(ret)
            self.finish()

    @authorization(roles=['admin'])
    async def delete(self, user_id):
        """
        Delete a user.

        Args:
            user_id (str): the user to delete

        Returns:
            dict: empty dict
        """
        await self.db.users.delete_one({'user_id':user_id})
        self.write({})
        self.finish()

class UserRolesHandler(AuthHandler):
    """
    Handle roles for an individual user.
    """
    @authorization(roles=['admin'])
    async def put(self, user_id):
        """
        Set the roles for a user.

        Body should contain {'roles': [ <role_name> ] }

        Args:
            user_id (str): the user

        Returns:
            dict: empty dict
        """
        data = json.loads(self.request.body)
        if 'roles' not in data:
            raise tornado.web.HTTPError(400, reason='missing roles')
        for role_name in data['roles']:
            ret = await self.db.roles.find_one({'name': role_name},
                    projection=['_id'])
            if not ret:
                raise tornado.web.HTTPError(400, reason='invalid role name')
        ret = await self.db.users.find_one_and_update({'user_id':user_id},
                {'$set': {'roles': data['roles']}})
        self.write({})
        self.finish()

class UserGroupsHandler(AuthHandler):
    """
    Handle groups for an individual user.
    """
    @authorization(roles=['admin','client'])
    async def get(self, user_id):
        """
        Get the groups for a user.

        Args:
            user_name (str): the user

        Returns:
            dict: {results: list of groups}
        """
        ret = await self.db.users.find_one({'user_id':user_id},
                projection={'_id':False})
        if not ret:
            self.send_error(404, reason="User not found")
        else:
            if 'groups' in ret:
                self.write({'results':ret['groups']})
            else:
                self.write({'results':[]})
            self.finish()

    @authorization(roles=['admin'])
    async def post(self, user_id):
        """
        Add a group to a user.

        Body should contain {'group': <group_name>}

        Args:
            user_id (str): the user

        Returns:
            dict: empty dict
        """
        data = json.loads(self.request.body)
        if 'group' not in data:
            raise tornado.web.HTTPError(400, reason='missing group')
        ret = await self.db.groups.find_one({'name': data['group']},
                projection=['_id'])
        if not ret:
            raise tornado.web.HTTPError(400, reason='invalid group')
        ret = await self.db.users.find_one_and_update({'user_id':user_id},
                {'$addToSet': {'groups': data['group']}})
        self.write({})
        self.finish()

    @authorization(roles=['admin'])
    async def put(self, user_id):
        """
        Set the groups for a user.

        Body should contain {'groups': [ <group_name> ] }

        Args:
            user_id (str): the user

        Returns:
            dict: empty dict
        """
        data = json.loads(self.request.body)
        if 'groups' not in data:
            raise tornado.web.HTTPError(400, 'missing groups')
        for g in data['groups']:
            ret = await self.db.groups.find_one({'name': g},
                    projection=['_id'])
            if not ret:
                raise tornado.web.HTTPError(400, 'invalid group')
        ret = await self.db.users.find_one_and_update({'user_id':user_id},
                {'$set': {'groups': data['groups']}})
        self.write({})
        self.finish()

class LDAPHandler(AuthHandler):
    """
    Handle LDAP authentication.
    """ 
    def initialize(self, ldap_uri=None, ldap_base=None, **kwargs):
        super(LDAPHandler, self).initialize(**kwargs)
        self.ldap_uri = ldap_uri
        self.ldap_base = ldap_base

    @catch_error
    async def post(self):
        """
        Validate LDAP login, creating a token.

        Body should contain {'username', 'password'}

        Returns:
            dict: {'token': auth token, 'roles':[roles], 'current_role':role}
        """
        try:
            data = json.loads(self.request.body)
            username = data['username']
            password = data['password']
            conn = ldap3.Connection(self.ldap_uri, 'uid={},{}'.format(username, self.ldap_base),
                                    password, auto_bind=True)
        except Exception:
            self.send_error(403, reason="Login failed")
        else:
            # get user info from DB
            ret = await self.db.users.find_one({'username':username})
            if not ret:
                # create generic user
                ret = {
                    'user_id': uuid.uuid1().hex,
                    'username': username,
                    'groups': ['users'],
                    'roles': ['user'],
                }
                await self.db.users.insert_one(ret)

            # create token
            token_data = {
                'username': username,
                'role': 'user' if 'user' in ret['roles'] else 'anonymous',
                'groups': ret['groups'],
            }
            tok = self.auth.create_token(username, type='user', payload=token_data)
            if isinstance(tok, bytes):
                tok = tok.decode('utf-8')
            self.write({
                'token': tok,
                'username': username,
                'roles': ret['roles'],
                'current_role': token_data['role'],
                'groups': token_data['groups'],
            })
            self.finish()

class CreateTokenHandler(AuthHandler):
    """
    Handle new token creation from an existing token.
    """
    @authorization(roles=['user','admin','client'])
    async def post(self):
        """
        Create a new token based on the existing token.

        Refresh is not allowed.  This is a sub-token with no
        extra expiration time.

        Body (json) args:
            type (str): the token type
            role (str): the role for the new token
            exp (int): expiration time (in seconds, from now)

        Returns:
            dict: {'result': token}
        """
        if 'username' not in self.auth_data:
            raise tornado.web.HTTPError(400, reason='invalid username')
        if 'role' not in self.auth_data:
            raise tornado.web.HTTPError(400, reason='invalid role')
        if 'type' not in self.auth_data or self.auth_data['type'] not in ('user','system'):
            raise tornado.web.HTTPError(400, reason='invalid token type')

        # get user info from DB
        ret = await self.db.users.find_one({'username':self.auth_data['username']})
        if not ret:
            raise tornado.web.HTTPError(400, reason='invalid username')

        # get args
        args = json.loads(self.request.body)

        # calculate expiration
        max_exp = self.auth_data['exp']-time.time()
        try:
            exp = int(args.get('exp', max_exp))
        except Exception:
            raise tornado.web.HTTPError(400, reason='invalid expiration')
        else:
            if exp < 0 or exp > max_exp:
                raise tornado.web.HTTPError(400, reason='invalid expiration')

        # create token
        data = {'username': self.auth_data['username']}
        tok_type = args.get('type', 'temp')
        logger.debug('token type: %r', tok_type)
        if tok_type == 'temp':
            # temp token for current role
            if self.auth_data['role'] not in ('admin','user'):
                raise tornado.web.HTTPError(400, reason='invalid role')
            if self.auth_data['role'] not in ret['roles']:
                raise tornado.web.HTTPError(400, reason='invalid role request')
            data.update({
                'role': self.auth_data['role'],
                'groups': ret['groups'] if 'groups' in ret else [],
            })
        elif tok_type == 'user':
            # switching roles
            if 'admin' not in ret['roles']:
                raise tornado.web.HTTPError(400, reason='invalid role')
            role = args.get('role', None)
            if (not role) or role not in ('admin','user'):
                raise tornado.web.HTTPError(400, reason='invalid role request')
            data.update({
                'role': role,
                'groups': ret['groups'] if 'groups' in ret else [],
            })
        elif tok_type == 'system':
            # iceprod internal tokens
            if self.auth_data['role'] not in ('admin','client'):
                raise tornado.web.HTTPError(400, reason='invalid role')
            role = args.get('role', None)
            if (not role) or not ((self.auth_data['role'] == 'admin' and role in ('system','client'))
                                  or (self.auth_data['role'] == 'client' and role == 'pilot')):
                raise tornado.web.HTTPError(400, reason='invalid role request')
            data.update({
                'role': role,
                'groups': [],
            })
        else:
            raise tornado.web.HTTPError(400, reason='invalid token type request')

        logger.debug('making new token with payload: %r', data)
        tok = self.auth.create_token(data['username'], expiration=exp,
                                     type=tok_type, payload=data)
        if isinstance(tok, bytes):
            tok = tok.decode('utf-8')
        self.write({'result': tok})

class AuthDatasetHandler(AuthHandler):
    """
    Handle dataset authorization rules.
    """
    @authorization(roles=['admin'])
    async def get(self, dataset_id):
        """
        Get the authorization rules for a dataset.

        Args:
            dataset_id (str): the dataset

        Returns:
            dict: {'read_groups': [ <group_id> ], 'write_groups': [ <group_id> ] }
        """
        ret = await self.db.auths_dataset.find_one({'dataset_id':dataset_id},
                projection={'_id':False,'dataset_id':False})
        if not ret:
            self.send_error(404, reason="Dataset not found")
        else:
            self.write(ret)
            self.finish()

    @authorization(roles=['admin','system']) # let other components (/datasets) set this
    async def put(self, dataset_id):
        """
        Set the authorization rules for a dataset.

        Body should contain `{'read_groups': [ <group_name> ], 'write_groups': [ <group_name> ] }`

        Args:
            dataset_id (str): the dataset

        Returns:
            dict: empty dict
        """
        data = json.loads(self.request.body)
        if 'read_groups' not in data:
            raise tornado.web.HTTPError(400, reason='missing read_groups')
        if 'write_groups' not in data:
            raise tornado.web.HTTPError(400, reason='missing write_groups')
        for group in set(data['read_groups'])|set(data['write_groups']):
            ret = await self.db.groups.find_one({'name': group},
                    projection=['_id'])
            if not ret:
                raise tornado.web.HTTPError(400, reason='invalid group name')
        ret = await self.db.auths_dataset.find_one_and_update({'dataset_id':dataset_id},
                {'$set': {'read_groups': data['read_groups'], 'write_groups': data['write_groups']}},
                upsert=True)
        self.write({})
        self.finish()

class AuthDatasetActionHandler(AuthHandler):
    """
    Handle dataset authorization rules.
    """
    @authenticated
    @catch_error
    async def get(self, dataset_id, action):
        """
        Check the auth token against the authorization rules for a dataset.

        Returns a 403 error on authorization failure.

        Args:
            dataset_id (str): the dataset
            action (str): 'read' or 'write'

        Returns:
            dict: empty dict
        """
        if action not in ('read','write'):
            raise tornado.web.HTTPError(400, reason='invalid action')
        if 'groups' not in self.auth_data:
            raise tornado.web.HTTPError(400, reason='invalid auth token')
        ret = await self.db.auths_dataset.find_one({'dataset_id':dataset_id},
                projection={'_id':False})
        if not ret:
            self.send_error(403, reason="Dataset not found")
        elif action+'_groups' not in ret:
            self.send_error(403, reason="Action not found")
        elif not set(ret[action+'_groups'])&set(self.auth_data['groups']):
            self.send_error(403, reason="Denied")
        else:
            self.write({})
            self.finish()



import logging
import json
import uuid
from collections import defaultdict

import tornado.web
import pymongo
import motor

from iceprod.server.rest import RESTHandler, RESTHandlerSetup, authorization
from iceprod.server.util import nowstr, job_statuses, job_status_sort

logger = logging.getLogger('rest.jobs')

def setup(config, *args, **kwargs):
    """
    Setup method for Jobs REST API.

    Sets up any database connections or other prerequisites.

    Args:
        config (dict): an instance of :py:class:`iceprod.server.config`.

    Returns:
        list: Routes for logs, which can be passed to :py:class:`tornado.web.Application`.
    """
    cfg_rest = config.get('rest',{}).get('jobs',{})
    db_cfg = cfg_rest.get('database',{})

    # add indexes
    db = pymongo.MongoClient(**db_cfg).jobs
    if 'job_id_index' not in db.jobs.index_information():
        db.jobs.create_index('job_id', name='job_id_index', unique=True)
    if 'dataset_id_index' not in db.jobs.index_information():
        db.jobs.create_index('dataset_id', name='dataset_id_index', unique=False)

    handler_cfg = RESTHandlerSetup(config, *args, **kwargs)
    handler_cfg.update({
        'database': motor.motor_tornado.MotorClient(**db_cfg).jobs,
    })

    return [
        (r'/jobs', MultiJobsHandler, handler_cfg),
        (r'/jobs/(?P<job_id>\w+)', JobsHandler, handler_cfg),
        (r'/datasets/(?P<dataset_id>\w+)/jobs', DatasetMultiJobsHandler, handler_cfg),
        (r'/datasets/(?P<dataset_id>\w+)/jobs/(?P<job_id>\w+)', DatasetJobsHandler, handler_cfg),
        (r'/datasets/(?P<dataset_id>\w+)/jobs/(?P<job_id>\w+)/status', DatasetJobsStatusHandler, handler_cfg),
        (r'/datasets/(?P<dataset_id>\w+)/job_actions/bulk_status/(?P<status>\w+)', DatasetJobBulkStatusHandler, handler_cfg),
        (r'/datasets/(?P<dataset_id>\w+)/job_summaries/status', DatasetJobSummariesStatusHandler, handler_cfg),
        (r'/datasets/(?P<dataset_id>\w+)/job_counts/status', DatasetJobCountsStatusHandler, handler_cfg),
    ]

class BaseHandler(RESTHandler):
    def initialize(self, database=None, **kwargs):
        super(BaseHandler, self).initialize(**kwargs)
        self.db = database

class MultiJobsHandler(BaseHandler):
    """
    Handle multi jobs requests.
    """
    @authorization(roles=['admin','system','client'])
    async def post(self):
        """
        Create a job entry.

        Body should contain the job data.

        Returns:
            dict: {'result': <job_id>}
        """
        data = json.loads(self.request.body)

        # validate first
        req_fields = {
            'dataset_id': str,
            'job_index': int,
        }
        for k in req_fields:
            if k not in data:
                raise tornado.web.HTTPError(400, reason='missing key: '+k)
            if not isinstance(data[k], req_fields[k]):
                r = 'key {} should be of type {}'.format(k, req_fields[k])
                raise tornado.web.HTTPError(400, reason=r)

        # validate job_id if given
        if 'job_id' in data:
            try:
                job_id = uuid.UUID(hex=data['job_id']).hex
            except Exception:
                raise tornado.web.HTTPError(400, reason='job_id should be a valid uuid')
        else:
            job_id = uuid.uuid1().hex

        # set some fields
        data.update({
            'job_id': job_id,
            'status': 'processing',
            'status_changed': nowstr(),
        })

        ret = await self.db.jobs.insert_one(data)
        self.set_status(201)
        self.write({'result': job_id})
        self.finish()

class JobsHandler(BaseHandler):
    """
    Handle single job requests.
    """
    @authorization(roles=['admin','client','system','pilot'])
    async def get(self, job_id):
        """
        Get a job entry.

        Args:
            job_id (str): the job id

        Returns:
            dict: job entry
        """
        ret = await self.db.jobs.find_one({'job_id':job_id},
                projection={'_id':False})
        if not ret:
            self.send_error(404, reason="Job not found")
        else:
            self.write(ret)
            self.finish()

    @authorization(roles=['admin','client','system'])
    async def patch(self, job_id):
        """
        Update a job entry.

        Body should contain the job data to update.  Note that this will
        perform a merge (not replace).

        Args:
            job_id (str): the job id

        Returns:
            dict: updated job entry
        """
        data = json.loads(self.request.body)
        if not data:
            raise tornado.web.HTTPError(400, reason='Missing update data')

        ret = await self.db.jobs.find_one_and_update({'job_id':job_id},
                {'$set':data},
                projection={'_id':False},
                return_document=pymongo.ReturnDocument.AFTER)
        if not ret:
            self.send_error(404, reason="Job not found")
        else:
            self.write(ret)
            self.finish()

class DatasetMultiJobsHandler(BaseHandler):
    """
    Handle multi jobs requests.
    """
    @authorization(roles=['admin','client','system'], attrs=['dataset_id:read'])
    async def get(self, dataset_id):
        """
        Get all jobs for a dataset.

        Params (optional):
            job_index: job_index to filter by
            status: | separated list of task status to filter by
            keys: | separated list of keys to return for each task

        Args:
            dataset_id (str): dataset id

        Returns:
            dict: {'job_id':{job_data}}
        """
        filters = {'dataset_id':dataset_id}
        status = self.get_argument('status', None)
        if status:
            filters['status'] = {'$in': status.split('|')}

        job_index = self.get_argument('job_index', None)
        if job_index:
            try:
                filters['job_index'] = int(job_index)
            except ValueError:
                raise tornado.web.HTTPError(400, reason='Bad argument "job_index": must be integer')

        projection = {'_id': False}
        keys = self.get_argument('keys','')
        if keys:
            projection.update({x:True for x in keys.split('|') if x})
            projection['job_id'] = True

        cursor = self.db.jobs.find(filters, projection=projection)
        ret = {}
        async for row in cursor:
            ret[row['job_id']] = row
        self.write(ret)
        self.finish()

class DatasetJobsHandler(BaseHandler):
    """
    Handle single job requests.
    """
    @authorization(roles=['admin','client','system'], attrs=['dataset_id:read'])
    async def get(self, dataset_id, job_id):
        """
        Get a job entry.

        Args:
            dataset_id (str): dataset id
            job_id (str): the job id

        Returns:
            dict: job entry
        """
        ret = await self.db.jobs.find_one({'job_id':job_id,'dataset_id':dataset_id},
                projection={'_id':False})
        if not ret:
            self.send_error(404, reason="Job not found")
        else:
            self.write(ret)
            self.finish()

class DatasetJobsStatusHandler(BaseHandler):
    """
    Handle single job requests.
    """
    @authorization(roles=['admin','client','system'], attrs=['dataset_id:write'])
    async def put(self, dataset_id, job_id):
        """
        Set a job status.

        Body should have {'status': <new_status>}

        Args:
            dataset_id (str): dataset id
            job_id (str): the job id

        Returns:
            dict: empty dict
        """
        data = json.loads(self.request.body)
        if (not data) or 'status' not in data:
            raise tornado.web.HTTPError(400, reason='Missing status in body')
        if data['status'] not in job_statuses:
            raise tornado.web.HTTPError(400, reason='Bad status')
        update_data = {
            'status': data['status'],
            'status_changed': nowstr(),
        }

        ret = await self.db.jobs.update_one({'job_id':job_id,'dataset_id':dataset_id},
                {'$set':update_data})
        if (not ret) or ret.modified_count < 1:
            self.send_error(404, reason="Job not found")
        else:
            self.write({})
            self.finish

class DatasetJobBulkStatusHandler(BaseHandler):
    """
    Update the status of multiple jobs at once.
    """
    @authorization(roles=['admin','client','system'], attrs=['dataset_id:write'])
    async def post(self, dataset_id, status):
        """
        Set multiple jobs' status.

        Body should have {'jobs': [<job_id>, <job_id>, ...]}

        Args:
            dataset_id (str): dataset id
            status (str): the status

        Returns:
            dict: empty dict
        """
        data = json.loads(self.request.body)
        if (not data) or 'jobs' not in data or not data['jobs']:
            raise tornado.web.HTTPError(400, reason='Missing jobs in body')
        jobs = list(data['jobs'])
        if len(jobs) > 100000:
            raise tornado.web.HTTPError(400, reason='Too many jobs specified (limit: 100k)')
        if status not in job_statuses:
            raise tornado.web.HTTPError(400, reason='Bad status')
        query = {
            'dataset_id': dataset_id,
            'job_id': {'$in': jobs},
        }
        update_data = {
            'status': status,
            'status_changed': nowstr(),
        }

        ret = await self.db.jobs.update_many(query, {'$set':update_data})
        if (not ret) or ret.modified_count < 1:
            self.send_error(404, reason="Jobs not found")
        else:
            self.write({})
            self.finish()

class DatasetJobSummariesStatusHandler(BaseHandler):
    """
    Handle job summary grouping by status.
    """
    @authorization(roles=['admin','client','system'], attrs=['dataset_id:read'])
    async def get(self, dataset_id):
        """
        Get the job summary for all jobs in a dataset, group by status.

        Args:
            dataset_id (str): dataset id

        Returns:
            dict: {<status>: [<job_id>,]}
        """
        cursor = self.db.jobs.find({'dataset_id':dataset_id},
                projection={'_id':False,'status':True,'job_id':True})
        ret = defaultdict(list)
        async for row in cursor:
            ret[row['status']].append(row['job_id'])
        ret2 = {}
        for k in sorted(ret, key=job_status_sort):
            ret2[k] = ret[k]
        self.write(ret2)
        self.finish()

class DatasetJobCountsStatusHandler(BaseHandler):
    """
    Handle job counts by status.
    """
    @authorization(roles=['admin','client','system'], attrs=['dataset_id:read'])
    async def get(self, dataset_id):
        """
        Get the job counts for all jobs in a dataset, group by status.

        Args:
            dataset_id (str): dataset id

        Returns:
            dict: {<status>: [<job_id>,]}
        """
        cursor = self.db.jobs.aggregate([
            {'$match':{'dataset_id':dataset_id}},
            {'$group':{'_id':'$status', 'total': {'$sum':1}}},
        ])
        ret = {}
        async for row in cursor:
            ret[row['_id']] = row['total']
        ret2 = {}
        for k in sorted(ret, key=job_status_sort):
            ret2[k] = ret[k]
        self.write(ret2)
        self.finish()



"""
Move tasks from reset to waiting/failed/suspended.

Initial delay: rand(5 minutes)
Periodic delay: 20 minutes
"""

import asyncio
import logging
import random
import time

from tornado.ioloop import IOLoop

logger = logging.getLogger('reset_tasks')

def reset_tasks(module):
    """
    Initial entrypoint.

    Args:
        module (:py:class:`iceprod.server.modules.schedule`): schedule module
    """
    # initial delay
    IOLoop.current().call_later(random.randint(60,60*5), run, module.rest_client)


async def reset_dataset(dataset_id, rest_client=None, debug=False):
    try:
        dataset = await rest_client.request('GET', '/datasets/{}'.format(dataset_id))
        tasks = await rest_client.request('GET', '/datasets/{}/task_summaries/status'.format(dataset_id))
        if 'reset' in tasks:
            logger.info('dataset %s reset tasks: %s', dataset_id, tasks['reset'])
            for task_id in tasks['reset']:
                try:
                    task = await rest_client.request('GET', '/tasks/{}'.format(task_id))
                    status = 'waiting'
                    if dataset['debug']:
                        status = 'suspended'
                    elif 'failures' in task and task['failures'] > 10:
                        status = 'failed'
                    args = {'status': status}
                    await rest_client.request('PUT', '/datasets/{}/tasks/{}/status'.format(dataset_id,task_id), args)
                except Exception:
                    logger.error('error resetting task %s', task_id, exc_info=True)
                    if debug:
                        raise
    except Exception:
        logger.error('error resetting tasks in dataset %s', dataset_id, exc_info=True)
        if debug:
            raise


async def run(rest_client, debug=False):
    """
    Actual runtime / loop.

    Args:
        rest_client (:py:class:`iceprod.core.rest_client.Client`): rest client
        debug (bool): debug flag to propagate exceptions
    """
    start_time = time.time()

    try:
        datasets = await rest_client.request('GET', '/dataset_summaries/status')
        dataset_ids = []
        if 'processing' in datasets:
            dataset_ids.extend(datasets['processing'])
        if 'truncated' in datasets:
            dataset_ids.extend(datasets['truncated'])
        awaitables = set()
        for dataset_id in dataset_ids:
            fut = asyncio.create_task(reset_dataset(dataset_id, rest_client=rest_client, debug=debug))
            awaitables.add(fut)
            if len(awaitables) >= 20:
                done, pending = await asyncio.wait(awaitables, return_when=asyncio.FIRST_COMPLETED)
                for fut in done:
                    await fut
                awaitables = pending
        for fut in asyncio.as_completed(awaitables):
            await fut
    except Exception:
        logger.error('error resetting non-active tasks', exc_info=True)
        if debug:
            raise

    # run again after 20 minute delay
    stop_time = time.time()
    delay = max(60*20 - (stop_time-start_time), 60)
    IOLoop.current().call_later(delay, run, rest_client)


def main():
    import argparse
    import asyncio
    import os
    parser = argparse.ArgumentParser(description='run a scheduled task once')
    parser.add_argument('-t', '--token', default=os.environ.get('ICEPROD_TOKEN', None), help='auth token')
    parser.add_argument('--dataset-id', default=None, help='dataset id to reset')
    parser.add_argument('--log-level', default='info', help='log level')
    parser.add_argument('--debug', default=False, action='store_true', help='debug enabled')

    args = parser.parse_args()
    args = vars(args)

    logformat='%(asctime)s %(levelname)s %(name)s %(module)s:%(lineno)s - %(message)s'
    logging.basicConfig(format=logformat, level=getattr(logging, args['log_level'].upper()))

    from rest_tools.client import RestClient
    rpc = RestClient('https://iceprod2-api.icecube.wisc.edu', args['token'])

    if args['dataset_id']:
        asyncio.run(reset_dataset(args['dataset_id'], rpc, debug=args['debug']))
    else:
        asyncio.run(run(rpc, debug=args['debug']))


if __name__ == '__main__':
    main()



"""
Clean up the pilots.

Initial delay: rand(60 minute)
Periodic delay: 60 minutes
"""

import logging
import random
import time
from datetime import datetime, timedelta

from tornado.ioloop import IOLoop

from iceprod.server.util import str2datetime

logger = logging.getLogger('pilot_monitor')

def pilot_cleanup(module):
    """
    Initial entrypoint.

    Args:
        module (:py:class:`iceprod.server.modules.schedule`): schedule module
    """
    # initial delay
    IOLoop.current().call_later(random.randint(60,60*60), run,
                                module.rest_client)

async def run(rest_client, debug=False):
    """
    Actual runtime / loop.

    Args:
        rest_client (:py:class:`iceprod.core.rest_client.Client`): rest client
        debug (bool): debug flag to propagate exceptions
    """
    start_time = time.time()
    time_limit = datetime.utcnow() - timedelta(days=14)
    async def reset_pilot(pilot_id):
        await rest_client.request('DELETE', '/pilots/{}'.format(pilot_id))
    try:
        pilots = await rest_client.request('GET', '/pilots')
        for pilot in pilots.values():
            if 'last_update' not in pilot or str2datetime(pilot['last_update']) < time_limit:
                await reset_pilot(pilot['pilot_id'])

    except Exception:
        logger.error('error cleaning pilots', exc_info=True)
        if debug:
            raise

    # run again after 5 minute delay
    stop_time = time.time()
    delay = max(60*60 - (stop_time-start_time), 60)
    IOLoop.current().call_later(delay, run, rest_client)


def main():
    import argparse
    import asyncio
    import os
    parser = argparse.ArgumentParser(description='run a scheduled task once')
    parser.add_argument('-t', '--token', default=os.environ.get('ICEPROD_TOKEN', None), help='auth token')
    parser.add_argument('--log-level', default='info', help='log level')
    parser.add_argument('--debug', default=False, action='store_true', help='debug enabled')

    args = parser.parse_args()
    args = vars(args)

    logformat='%(asctime)s %(levelname)s %(name)s %(module)s:%(lineno)s - %(message)s'
    logging.basicConfig(format=logformat, level=getattr(logging, args['log_level'].upper()))

    from rest_tools.client import RestClient
    rpc = RestClient('https://iceprod2-api.icecube.wisc.edu', args['token'])

    asyncio.run(run(rpc, debug=args['debug']))

if __name__ == '__main__':
    main()



"""
gridftp interface
"""

from __future__ import absolute_import, division, print_function

import os
import logging
from threading import Thread, Event
from functools import partial
from collections import namedtuple
from datetime import datetime
import time
import tempfile
import shutil
import subprocess

logger = logging.getLogger('gridftp')

def _cmd(cmd, timeout=1200):
    subprocess.run(cmd, timeout=timeout, check=True)

def _cmd_output(cmd, timeout=1200):
    p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    try:
        output = p.communicate(timeout=timeout)[0].decode('utf-8')
        return (p.returncode, output)
    except subprocess.TimeoutExpired:
        p.kill()
        raise Exception('Request timed out')


def listify(lines,details=False,dotfiles=False):
    """Turn ls output into a list of NamedTuples"""
    out = []
    if details:
        File = namedtuple('File', ['directory','perms','subfiles',
                                   'owner','group','size','date',
                                   'name'])
        months = {'jan':1,'feb':2,'mar':3,'apr':4,'may':5,'jun':6,
                  'jul':7,'aug':8,'sep':9,'oct':10,'nov':11,'dec':12}
        for x in lines.split('\n'):
            if not x.strip():
                continue
            pieces = x.split()
            name = os.path.basename(pieces[-1])
            if name.startswith('.') and not dotfiles:
                continue
            d = x[0] == 'd'
            perms = pieces[0][1:]
            subfiles = 0
            if pieces[1].isdigit():
                subfiles = int(pieces[1])
                pieces = pieces[1:]
            else:
                subfiles = int(pieces[3])
            year = datetime.now().year
            month = months[pieces[4].lower()]
            day = int(pieces[5])
            if ':' in pieces[6]:
                hour,minute = pieces[6].split(':')
                dt = datetime(year,month,day,int(hour),int(minute))
            else:
                year = int(pieces[6])
                dt = datetime(year,month,day)
            out.append(File(d,perms,subfiles,pieces[1],pieces[2],
                            int(pieces[3]),dt,name))
    else:
        for x in lines.split('\n'):
            if not x.strip():
                continue
            f = x.split()[-1]
            if not f.startswith('.') or dotfiles:
                out.append(f)
    return out

class GridFTP(object):
    """GridFTP interface to command line client.

       Example:
           GridFTP.get('gsiftp://data.icecube.wisc.edu/file',
                       filename='/path/to/file')
    """

    _timeout = 3600 # 1 hour default timeout

    @classmethod
    def supported_address(cls,address):
        """Return False for address types that are not supported"""
        if '://' not in address:
            return False
        addr_type = address.split(':')[0]
        if addr_type not in ('gsiftp','ftp'):
            return False
        return True

    @classmethod
    def address_split(cls,address):
        """Split an address into server/path parts"""
        pieces = address.split('://',1)
        if '/' in pieces[1]:
            pieces2 = pieces[1].split('/',1)
            return (pieces[0]+'://'+pieces2[0],'/'+pieces2[1])
        else:
            return (address,'/')

    @classmethod
    def get(cls, address, filename=None, request_timeout=None):
        """
        Do a GridFTP get request.

        Either data is returned directly or filename must be defined.

        Args:
            address (str): url to get from
            filename (str): filename to write data to
            request_timeout (float): timeout in secodns

        Returns:
            str: data, if filename is not defined

        Raises:
            Exception for failure
        """
        if not cls.supported_address(address):
            raise Exception('address type not supported for address %s'%str(address))

        tmpdir = None
        if filename is None:
            tmpdir = tempfile.mkdtemp(dir=os.getcwd())
            dest = 'file:'+os.path.join(tmpdir,'get_tmp_file')
        else:
            dest = 'file:'+filename

        cmd = ['globus-url-copy',address,dest]

        if request_timeout is None:
            timeout = cls._timeout
        else:
            timeout = request_timeout

        try:
            _cmd(cmd, timeout=timeout)
            if filename is None:
                with open(dest[5:]) as f:
                    return f.read()
        finally:
            if tmpdir:
                shutil.rmtree(tmpdir,ignore_errors=True)

    @classmethod
    def put(cls, address, data=None, filename=None, request_timeout=None):
        """
        Do a GridFTP put request.

        Either data or filename must be defined.

        Args:
            address (str): url to put to
            data (str): the data to put
            filename (str): filename for data to put
            request_timeout (float): timeout in seconds

        Raises:
            Exception for failure
        """
        if not cls.supported_address(address):
            raise Exception('address type not supported for address %s'%str(address))

        tmpdir = None
        if data is not None:
            tmpdir = tempfile.mkdtemp(dir=os.getcwd())
            src = 'file:'+os.path.join(tmpdir,'put_tmp_file')
            with open(src[5:],'w' if isinstance(data,str) else 'wb') as f:
                f.write(data)
        elif filename is not None:
            src = 'file:'+filename
        else:
            raise Exception('Neither data or filename is defined')

        cmd = ['globus-url-copy','-cd',src,address]

        if request_timeout is None:
            timeout = cls._timeout
        else:
            timeout = request_timeout

        try:
            _cmd(cmd, timeout=timeout)
        finally:
            if tmpdir:
                shutil.rmtree(tmpdir,ignore_errors=True)

    @classmethod
    def list(cls, address, request_timeout=None, details=False, dotfiles=False):
        """
        Do a GridFTP list request.


        Args:
            address (str): url to list
            request_timeout (float): timeout in seconds
            details (bool): result is a list of NamedTuples
            dotfiles (bool): result includes '.', '..', and other '.' files

        Returns:
           list: a list of files

        Raises:
            Exception on error
        """
        if not cls.supported_address(address):
            raise Exception('address type not supported for address %s'%str(address))

        cmd = ['uberftp','-retry','5','-ls',address]

        if request_timeout is None:
            timeout = cls._timeout
        else:
            timeout = request_timeout

        ret = _cmd_output(cmd, timeout=timeout)
        if ret[0]:
            logger.warning(f'{ret[1]}')
            raise Exception('Error getting listing')
        return listify(ret[1], details=details, dotfiles=dotfiles)

    @classmethod
    def mkdir(cls, address, request_timeout=None, parents=False):
        """
        Make a directory on the ftp server.

        Args:
            address (str): url to directory
            request_timeout (float): timeout in seconds
            parents (bool): make parent directories as needed

        Raises:
            Exception on error
        """
        if not cls.supported_address(address):
            raise Exception('address type not supported for address %s'%str(address))

        if parents:
            # recursively make directory
            try:
                cls.mkdir(os.path.basename(address),
                          request_timeout=request_timeout,parents=True)
            except Exception:
                pass

        cmd = ['uberftp','-retry','5','-mkdir',address]

        if request_timeout is None:
            timeout = cls._timeout
        else:
            timeout = request_timeout

        _cmd(cmd, timeout=timeout)

    @classmethod
    def rmdir(cls, address, request_timeout=None):
        """
        Remove a directory on the ftp server.

        This fails if the directory is not empty.  Use :py:func:`rmtree` for
        recursive removal.

        Args:
            address (str): url to directory
            request_timeout (float): timeout in seconds

        Raises:
            Exception on error
        """
        if not cls.supported_address(address):
            raise Exception('address type not supported for address %s'%str(address))

        cmd = ['uberftp','-retry','5','-rmdir',address]

        if request_timeout is None:
            timeout = cls._timeout
        else:
            timeout = request_timeout

        ret = _cmd_output(cmd, timeout=timeout)
        if ret[0] and 'No match for' not in ret[1]:
            raise Exception('Error removing dir')

    @classmethod
    def delete(cls, address, request_timeout=None):
        """
        Delete a file on the ftp server.

        Args:
            address (str): url to file
            request_timeout (float): timeout in seconds

        Raises:
            Exception on error
        """
        if not cls.supported_address(address):
            raise Exception('address type not supported for address %s'%str(address))

        cmd = ['uberftp','-retry','5','-rm',address]

        if request_timeout is None:
            timeout = cls._timeout
        else:
            timeout = request_timeout

        ret = _cmd_output(cmd, timeout=timeout)
        if ret[0] and 'No match for' not in ret[1]:
            raise Exception('Error removing dir')

    @classmethod
    def rmtree(cls, address, request_timeout=None):
        """
        Delete a file or directory on the ftp server.

        This is recursive, like `rm -rf`.

        Args:
            address (str): url to file or directory
            request_timeout (float): timeout in seconds

        Raises:
            Exception on error
        """
        if not cls.supported_address(address):
            raise Exception('address type not supported for address %s'%str(address))

        cmd = ['uberftp','-retry','5','-rm','-r',address]

        if request_timeout is None:
            timeout = cls._timeout
        else:
            timeout = request_timeout

        ret = _cmd_output(cmd, timeout=timeout)
        if ret[0] and 'No match for' not in ret[1]:
            raise Exception('Error removing dir')

    @classmethod
    def move(cls, src, dest, request_timeout=None):
        """
        Move a file on the ftp server.

        Args:
            src (str): url to source file
            dest (str): url to destination file
            request_timeout (float): timeout in seconds

        Raises:
            Exception on error
        """
        if not cls.supported_address(src):
            raise Exception('address type not supported for src %s'%str(src))
        if not cls.supported_address(dest):
            raise Exception('address type not supported for dest %s'%str(dest))

        cmd = ['uberftp','-retry','5','-rename',src,cls.address_split(dest)[-1]]

        if request_timeout is None:
            timeout = cls._timeout
        else:
            timeout = request_timeout

        _cmd(cmd, timeout=timeout)

    @classmethod
    def exists(cls, address, request_timeout=None):
        """
        Check if a file exists on the ftp server.

        Args:
            address (str): url to file
            request_timeout (float): timeout in seconds

        Raises:
            Exception on error
        """
        if not cls.supported_address(address):
            raise Exception('address type not supported for address %s'%str(address))

        cmd = ['uberftp','-retry','5','-size',address]

        if request_timeout is None:
            timeout = cls._timeout
        else:
            timeout = request_timeout

        ret = _cmd_output(cmd, timeout=timeout)
        return (not ret[0])

    @classmethod
    def chmod(cls, address, mode, request_timeout=None):
        """
        Chmod a file on the ftp server.

        Args:
            address (str): url to file
            mode (str): mode of file
            request_timeout (float): timeout in seconds

        Raises:
            Exception on error
        """
        if not cls.supported_address(address):
            raise Exception('address type not supported for address %s'%str(address))

        cmd = ['uberftp','-retry','5','-chmod',mode,address]

        if request_timeout is None:
            timeout = cls._timeout
        else:
            timeout = request_timeout

        _cmd(cmd, timeout=timeout)

    @classmethod
    def size(cls, address, request_timeout=None):
        """
        Get the size of a file on the ftp server.

        Args:
            address (str): url to file
            request_timeout (float): timeout in seconds

        Returns:
            int: size of file in bytes

        Raises:
            Exception on error
        """
        if not cls.supported_address(address):
            raise Exception('address type not supported for address %s'%str(address))

        cmd = ['uberftp','-retry','5','-size',address]

        if request_timeout is None:
            timeout = cls._timeout
        else:
            timeout = request_timeout

        ret = _cmd_output(cmd, timeout=timeout)
        if ret[0]:
            raise Exception('failed to get size')
        return int(ret[1])

    @classmethod
    def _chksum(cls, type, address, request_timeout=None):
        """Chksum is faked by redownloading the file and checksumming that"""
        from iceprod.core.functions import cksm
        if not cls.supported_address(address):
            raise Exception('address type not supported for address %s'%str(address))
        if type.endswith('sum'):
            type = type[:-3]

        tmpdir = tempfile.mkdtemp(dir=os.getcwd())
        dest = 'file:'+os.path.join(tmpdir,'dest')

        cmd = ['globus-url-copy',address,dest]

        if request_timeout is None:
            timeout = cls._timeout
        else:
            timeout = request_timeout

        try:
            _cmd(cmd, timeout=timeout)
            if not os.path.exists(dest[5:]):
                raise Exception('failed to redownload')
            return cksm(dest[5:],type)
        finally:
            shutil.rmtree(tmpdir,ignore_errors=True)

    ### Some helper functions for different checksum types ###

    @classmethod
    def md5sum(cls,address,request_timeout=None):
        """
        Get the md5sum of a file on an ftp server.

        Args:
            address (str): url to file
            request_timeout (float): timeout in seconds

        Returns:
            str: the md5sum

        Raises:
            Exception on error
        """
        return cls._chksum('md5sum',address,request_timeout=request_timeout)

    @classmethod
    def sha1sum(cls,address,request_timeout=None):
        """
        Get the sha1sum of a file on an ftp server.

        Args:
            address (str): url to file
            request_timeout (float): timeout in seconds

        Returns:
            str: the sha1sum

        Raises:
            Exception on error
        """
        return cls._chksum('sha1sum',address,request_timeout=request_timeout)

    @classmethod
    def sha256sum(cls, address, request_timeout=None):
        """
        Get the sha256sum of a file on an ftp server.

        Args:
            address (str): url to file
            request_timeout (float): timeout in seconds

        Returns:
            str: the sha256sum

        Raises:
            Exception on error
        """
        return cls._chksum('sha256sum',address,request_timeout=request_timeout)

    @classmethod
    def sha512sum(cls, address, request_timeout=None):
        """
        Get the sha512sum of a file on an ftp server.

        Args:
            address (str): url to file
            request_timeout (float): timeout in seconds

        Returns:
            str: the sha512sum

        Raises:
            Exception on error
        """
        return cls._chksum('sha512sum',address,request_timeout=request_timeout)
